import numpy as np
import pandas as pd
import re
from fuzzywuzzy import fuzz
import zipfile
import os
from .text_functions import STOPWORDS
from . import home
from .utils import download_file

zip_location = home+'/rules-based.zip'

if not os.path.isfile(zip_location):
    print('downloading ZIP rules-based')
    download_file("rules-based.zip", zip_location)
    with zipfile.ZipFile(zip_location, 'r') as zip_ref:
        zip_ref.extractall(home)

def apply_stopwords_calon(string):
    string = re.sub('[^A-Za-z ]+', '', string)
    return ' '.join([i for i in string.split() if i not in STOPWORDS and len(i) > 1])

df = pd.read_csv(home+'/rules-based/calon.csv')
namacalon = df.NamaCalon.str.lower().unique().tolist()
for i in range(len(namacalon)):
    namacalon[i] = apply_stopwords_calon(namacalon[i])

df = pd.read_csv(home+'/rules-based/negeri.csv')
negeri = df.negeri.str.lower().unique().tolist()
parlimen = df.parlimen.str.lower().unique().tolist()
dun = df.dun.str.lower().unique().tolist()[:-1]

with open(home+'/rules-based/person-normalized','r') as fopen:
    person = list(filter(None, fopen.read().split('\n')))

person_dict = {}
for i in range(len(person)):
    splitted = person[i].split(':')
    uniques = list(filter(None,(set([k.strip().lower() for k in splitted[1].split(', ')] + [splitted[0].lower()]))))
    person_dict[splitted[0]] = uniques

with open(home+'/rules-based/topic-normalized','r') as fopen:
    topic = list(filter(None, fopen.read().split('\n')))

topic_dict = {}
for i in range(len(topic)):
    splitted = topic[i].split(':')
    uniques = list(filter(None,(set([k.strip().lower() for k in splitted[1].split(', ')] + [splitted[0].lower()]))))
    topic_dict[splitted[0]] = uniques

with open(home+'/rules-based/short-normalized','r') as fopen:
    short = list(filter(None, fopen.read().split('\n')))

short_dict = {}
for i in range(len(short)):
    splitted = short[i].split(':')
    uniques = list(filter(None,(set([k.strip().lower() for k in splitted[1].split(', ')] + [splitted[0].lower()]))))
    short_dict[splitted[0]] = uniques

def get_influencers(string):
    assert (isinstance(string, str)), "input must be a string"
    string = string.lower()
    influencers = []
    for key, vals in person_dict.items():
        for v in vals:
            if string.find(v) >= 0:
                influencers.append(key)
                break
    for key, vals in short_dict.items():
        for v in vals:
            if v in string.split():
                influencers.append(key)
                break

    for index in np.where(np.array([fuzz.token_set_ratio(i, string) for i in namacalon]) >= 80)[0]:
        influencers.append(namacalon[index])
    return list(set(influencers))

def get_topics(string):
    assert (isinstance(string, str)), "input must be a string"
    string = string.lower()
    topics = []
    for key, vals in topic_dict.items():
        for v in vals:
            if string.find(v) >= 0:
                topics.append(key)
                break
    for key, vals in person_dict.items():
        for v in vals:
            if string.find(v) >= 0:
                topics.append(key)
                break
    for key, vals in short_dict.items():
        for v in vals:
            if v in string.split():
                topics.append(key)
                break

    topics += [i for i in negeri if i in string.split()]
    topics += [i for i in parlimen if i in string.split()]
    topics += [i for i in dun if i in string.split()]
    return list(set(topics))



from __future__ import print_function, unicode_literals

BASE = {0: [],
        1: ["satu"],
        2: ["dua"],
        3: ["tiga"],
        4: ["empat"],
        5: ["lima"],
        6: ["enam"],
        7: ["tujuh"],
        8: ["lapan"],
        9: ["sembilan"]}

TENS_TO = {3: "ribu",
           6: "juta",
           9: "billion",
           12: "trillion",
           15: "quadrillion",
           18: "quintillion",
           21: "sextillion",
           24: "septillion",
           27: "oktillion",
           30: "nonillion",
           33: "decillion"}

errmsg_floatord = "Cannot treat float number as ordinal"
errmsg_negord = "Cannot treat negative number as ordinal"
errmsg_toobig = "Too large"
max_num = 10 ** 36

def verify_ordinal(value):
    if not value == int(value):
        raise TypeError(errmsg_floatord % value)
    if not abs(value) == value:
        raise TypeError(errmsg_negord % value)

def split_by_koma(number):
    return str(number).split('.')

def ratus(number):
    if number == '1':
        return ['seratus']
    elif number == '0':
        return []
    else:
        return BASE[int(number)] + ['ratus']

def puluh(number):
    if number[0] == '1':
        if number[1] == '0':
            return ['sepuluh']
        elif number[1] == '1':
            return ['sebelas']
        else:
            return BASE[int(number[1])] + ['belas']
    elif number[0] == '0':
        return BASE[int(number[1])]
    else:
        return (BASE[int(number[0])] + ['puluh']+BASE[int(number[1])])

def split_by_3(number):
    blocks = ()
    length = len(number)
    if length < 3:
        blocks += ((number,),)
    else:
        len_of_first_block = length % 3
        if len_of_first_block > 0:
            blocks += number[0:len_of_first_block],
        for i in range(len_of_first_block, length, 3):
            blocks += (number[i:i + 3],),
    return blocks

def spell(blocks):
    word_blocks = ()
    first_block = blocks[0]
    if len(first_block[0]) == 1:
        if first_block[0] == '0':
            spelling = ['nol']
        else:
            spelling = BASE[int(first_block[0])]
    elif len(first_block[0]) == 2:
        spelling = puluh(first_block[0])
    else:
        spelling = (ratus(first_block[0][0]) + puluh(first_block[0][1:3]))
    word_blocks += (first_block[0], spelling),
    for block in blocks[1:]:
        spelling = ratus(block[0][0]) + puluh(block[0][1:3])
        block += spelling,
        word_blocks += block,
    return word_blocks

def spell_float(float_part):
    word_list = []
    for n in float_part:
        if n == '0':
            word_list += ['nol']
            continue
        word_list += BASE[int(n)]
    return ' '.join(['', 'perpuluhan'] + word_list)

def join(word_blocks, float_part):
    word_list = []
    length = len(word_blocks) - 1
    first_block = word_blocks[0],
    start = 0

    if length == 1 and first_block[0][0] == '1':
        word_list += ['seribu']
        start = 1

    for i in range(start, length + 1, 1):
        word_list += word_blocks[i][1]
        if not word_blocks[i][1]:
            continue
        if i == length:
            break
        word_list += [TENS_TO[(length - i) * 3]]

    return ' '.join(word_list) + float_part

def to_cardinal(number):
    if number >= max_num:
        raise OverflowError(errmsg_toobig % (number, max_num))
    minus = ''
    if number < 0:
        minus = 'negatif '
    float_word = ''
    n = split_by_koma(abs(number))
    if len(n) == 2:
        float_word = spell_float(n[1])
    return minus + join(spell(split_by_3(n[0])), float_word)

def to_ordinal(number):
    verify_ordinal(number)
    out_word = to_cardinal(number)
    if out_word == "satu":
        return "pertama"
    return "ke" + out_word

def to_ordinal_num(number):
    verify_ordinal(number)
    return "ke-" + str(number)

def to_currency(value):
    return to_cardinal(value) + " ringgit"

def to_year(value):
    return to_cardinal(value)



import re
import os
import numpy as np
from nltk.tokenize import word_tokenize
from unidecode import unidecode
import itertools
from .utils import download_file
from .tatabahasa import stopword_tatabahasa
from . import home

STOPWORDS = None
stopwords_location = home+'/stop-word-kerulnet'
if not os.path.isfile(stopwords_location):
    print('downloading stopwords')
    download_file("stop-word-kerulnet", stopwords_location)
with open(stopwords_location,'r') as fopen:
    STOPWORDS = list(filter(None, fopen.read().split('\n')))

STOPWORDS = list(set(STOPWORDS + stopword_tatabahasa))
VOWELS = "aeiou"
PHONES = ['sh', 'ch', 'ph', 'sz', 'cz', 'sch', 'rz', 'dz']

def isWord(word):
    if word:
        consecutiveVowels = 0
        consecutiveConsonents = 0
        for idx, letter in enumerate(word.lower()):
            vowel = True if letter in VOWELS else False
            if idx:
                prev = word[idx-1]
                prevVowel = True if prev in VOWELS else False
                if not vowel and letter == 'y' and not prevVowel:
                    vowel = True
                if prevVowel != vowel:
                    consecutiveVowels = 0
                    consecutiveConsonents = 0
            if vowel:
                consecutiveVowels += 1
            else:
                consecutiveConsonents +=1
            if consecutiveVowels >= 3 or consecutiveConsonents > 3:
                return False
            if consecutiveConsonents == 3:
                subStr = word[idx-2:idx+1]
                if any(phone in subStr for phone in PHONES):
                    consecutiveConsonents -= 1
                    continue
                return False
    return True

list_laughing = ['huhu','haha','gaga','hihi','wkawka','wkwk','kiki','keke','huehue']
def malaya_textcleaning(string):
    string = re.sub('http\S+|www.\S+', '',' '.join([i for i in string.split() if i.find('#')<0 and i.find('@')<0]))
    string = unidecode(string).replace('.', '. ')
    string = string.replace(',', ', ')
    string = re.sub('[^\'\"A-Za-z\- ]+', ' ', string)
    string = [y.strip() for y in word_tokenize(string.lower()) if isWord(y.strip())]
    string = [y for y in string if all([y.find(k) < 0 for k in list_laughing]) and y[:len(y)//2] != y[len(y)//2:]]
    string = ' '.join(string)
    string = (''.join(''.join(s)[:2] for _, s in itertools.groupby(string))).split()
    return ' '.join([y for y in string if y not in STOPWORDS])

def normalizer_textcleaning(string):
    string = re.sub('http\S+|www.\S+', '',' '.join([i for i in string.split() if i.find('#')<0 and i.find('@')<0]))
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = [y.strip() for y in string.lower().split() if len(y.strip())]
    string = [y for y in string if all([y.find(k) < 0 for k in list_laughing])]
    string = ' '.join(string).lower()
    return ''.join(''.join(s)[:2] for _, s in itertools.groupby(string))

def simple_textcleaning(string, decode=False):
    '''
    use by language_detection, topic modelling
    only accept A-Z, a-z
    '''
    string = unidecode(string) if decode else string
    string = re.sub('[^A-Za-z ]+', ' ', string)
    string = [y.strip() for y in string if len(y)]
    return ' '.join(string).lower()

def entities_textcleaning(string):
    string = re.sub('[^A-Za-z0-9\-\/ ]+', ' ', string).split()
    return [y.strip() for y in string if len(y)]

def summary_textcleaning(string):
    string = re.sub('[^A-Za-z0-9\-\/\'\"\.\, ]+', ' ', string).split()
    return ' '.join([y.strip() for y in string if len(y)])

def classification_textcleaning(string,no_stopwords=False):
    '''
    use by our text classifiers, stemmer, summarization, topic-modelling
    remove links, hashtags, alias
    '''
    string = re.sub('http\S+|www.\S+', '',' '.join([i for i in string.split() if i.find('#')<0 and i.find('@')<0]))
    string = unidecode(string).replace('.', '. ').replace(',', ', ')
    string = re.sub('[^A-Za-z\- ]+', ' ', string)
    if no_stopwords:
        return ' '.join([i for i in re.findall("[\\w']+|[;:\-\(\)&.,!?\"]", string) if len(i)>1]).lower()
    else:
        return ' '.join([i for i in re.findall("[\\w']+|[;:\-\(\)&.,!?\"]", string) if len(i)>1 and i not in STOPWORDS]).lower()

def process_word_pos_entities(word):
    word = word.lower()
    word = re.sub('[^A-Za-z0-9\- ]+', '', word)
    if word.isdigit():
        word = 'NUM'
    return word

def separate_dataset(trainset):
    datastring = []
    datatarget = []
    for i in range(len(trainset.data)):
        data_ = trainset.data[i].split('\n')
        data_ = list(filter(None, data_))
        datastring += data_
        for n in range(len(data_)):
            datatarget.append(trainset.target[i])
    return datastring, datatarget

def print_topics_modelling(topics, feature_names, sorting, topics_per_chunk=6, n_words=20):
    for i in range(0, len(topics), topics_per_chunk):
        these_topics = topics[i: i + topics_per_chunk]
        len_this_chunk = len(these_topics)
        print(("topic {:<8}" * len_this_chunk).format(*these_topics))
        print(("-------- {0:<5}" * len_this_chunk).format(""))
        for i in range(n_words):
            print(("{:<14}" * len_this_chunk).format(*feature_names[sorting[these_topics, i]]))
        print("\n")

def cluster_words(list_words):
    dict_words = {}
    for word in list_words:
        found = False
        for key in dict_words.keys():
            if word in key or any([word in inside for inside in dict_words[key]]):
                dict_words[key].append(word)
                found = True
            if key in word:
                dict_words[key].append(word)
        if not found:
            dict_words[word] = [word]
    results = []
    for key, words in dict_words.items():
        results.append(max(list(set([key] + words)), key=len))
    return list(set(results))

def str_idx(corpus, dic, maxlen, UNK=0):
    X = np.zeros((len(corpus),maxlen))
    for i in range(len(corpus)):
        for no, k in enumerate(corpus[i].split()[:maxlen][::-1]):
            try:
                X[i,-1 - no]=dic[k]
            except Exception as e:
                X[i,-1 - no]=UNK
    return X

def stemmer_str_idx(corpus, dic, UNK=3):
    X = []
    for i in corpus:
        ints = []
        for k in i:
            try:
                ints.append(dic[k])
            except Exception as e:
                ints.append(UNK)
        X.append(ints)
    return X

def fasttext_str_idx(texts, dictionary):
    idx_trainset = []
    for text in texts:
        idx = []
        for t in text.split():
            try:
                idx.append(dictionary[t])
            except:
                pass
        idx_trainset.append(idx)
    return idx_trainset

def pad_sentence_batch(sentence_batch, pad_int):
    padded_seqs = []
    seq_lens = []
    max_sentence_len = max([len(sentence) for sentence in sentence_batch])
    for sentence in sentence_batch:
        padded_seqs.append(sentence + [pad_int] * (max_sentence_len - len(sentence)))
        seq_lens.append(len(sentence))
    return padded_seqs, seq_lens

def add_ngram(sequences, token_indice):
    new_sequences = []
    for input_list in sequences:
        new_list = input_list[:]
        for ngram_value in range(2, 3):
            for i in range(len(new_list) - ngram_value + 1):
                ngram = tuple(new_list[i:i + ngram_value])
                if ngram in token_indice:
                    new_list.append(token_indice[ngram])
        new_sequences.append(new_list)
    return new_sequences

def char_str_idx(corpus, dic, UNK=0):
    maxlen = max([len(i) for i in corpus])
    X = np.zeros((len(corpus),maxlen))
    for i in range(len(corpus)):
        for no, k in enumerate(corpus[i][:maxlen][::-1]):
            try:
                X[i,-1 - no]=dic[k]
            except Exception as e:
                X[i,-1 - no]=UNK
    return X

def generate_char_seq(batch,idx2word,char2idx):
    x = [[len(idx2word[i]) for i in k] for k in batch]
    maxlen = max([j for i in x for j in i])
    temp = np.zeros((batch.shape[0],batch.shape[1],maxlen),dtype=np.int32)
    for i in range(batch.shape[0]):
        for k in range(batch.shape[1]):
            for no, c in enumerate(idx2word[batch[i,k]].lower()):
                temp[i,k,-1-no] = char2idx[c]
    return temp



import re
import os
import json
import tensorflow as tf
from tensorflow.contrib.seq2seq.python.ops import beam_search_ops
from Sastrawi.Stemmer.StemmerFactory import StemmerFactory
from .tatabahasa import permulaan, hujung
from .utils import load_graph, download_file
from .text_functions import pad_sentence_batch, stemmer_str_idx, classification_textcleaning
from . import home

factory = StemmerFactory()
sastrawi = factory.create_stemmer()
stemmer_json = home+'/stemmer-deep.json'
stemmer_graph = home+'/stemmer-deep.pb'
GO = 0
PAD = 1
EOS = 2
UNK = 3

class DEEP_STEMMER:
    def __init__(self, x, logits, sess, dicts):
        self._sess = sess
        self._x = x
        self._logits = logits
        self._dicts = dicts
        self._dicts['rev_dictionary_to'] = {int(k):v for k,v in self._dicts['rev_dictionary_to'].items()}

    def stem(self,string):
        assert (isinstance(string, str)), "input must be a string"
        token_strings = classification_textcleaning(string,True).split()
        idx = stemmer_str_idx(token_strings,self._dicts['dictionary_from'])
        predicted = self._sess.run(self._logits, feed_dict={self._x:pad_sentence_batch(idx, PAD)[0]})
        results = []
        for word in predicted:
            results.append(''.join([self._dicts['rev_dictionary_to'][c] for c in word if c not in[GO,PAD,EOS,UNK]]))
        return ' '.join(results)

def naive_stemmer(word):
    assert (isinstance(word, str)), "input must be a string"
    try:
        word = re.findall(r'^(.*?)(%s)$'%('|'.join(hujung)), word)[0][0]
        mula = re.findall(r'^(.*?)(%s)'%('|'.join(permulaan[::-1])), word)[0][1]
        return word.replace(mula,'')
    except:
        return word

def sastrawi_stemmer(string):
    assert (isinstance(string, str)), "input must be a string"
    return sastrawi.stem(string)

def deep_stemmer():
    if not os.path.isfile(stemmer_json):
        print('downloading JSON stemmer')
        download_file("v5/stemmer-deep.json", stemmer_json)
    with open(stemmer_json,'r') as fopen:
        dic_stemmer = json.load(fopen)
    if not os.path.isfile(stemmer_graph):
        print('downloading stemmer graph')
        download_file("v5/stemmer-frozen-model.pb", stemmer_graph)
    g=load_graph(stemmer_graph)
    return DEEP_STEMMER(g.get_tensor_by_name('import/Placeholder:0'),
                        g.get_tensor_by_name('import/logits:0'),
                        tf.InteractiveSession(graph=g),
                        dic_stemmer)



import pickle
import os
import xgboost as xgb
import numpy as np
from .text_functions import simple_textcleaning
from .utils import download_file
from . import home

bow_pkl = home+'/bow-language-detection.pkl'
multinomial_pkl = home+'/multinomial-language-detection.pkl'

xgb_bow_pkl = home+'/bow-xgb-language-detection.pkl'
xgb_pkl = home+'/xgb-language-detection.pkl'

lang_labels = {0: 'OTHER',1: 'ENGLISH',2: 'INDONESIA',3: 'MALAY'}

def get_language_labels():
    return lang_labels

class USER_XGB:
    def __init__(self, xgb, label, vectorize):
        self.xgb = xgb
        self.label = label
        self.vectorize = vectorize
    def predict(self, string, get_proba=False):
        assert (isinstance(string, str)), "input must be a string"
        vectors = self.vectorize.transform([simple_textcleaning(string,True)])
        result = self.xgb.predict(xgb.DMatrix(vectors),ntree_limit=self.xgb.best_ntree_limit)[0]
        if get_proba:
            return {self.label[i]:result[i] for i in range(len(result))}
        else:
            return self.label[np.argmax(result)]
    def predict_batch(self, strings, get_proba=False):
        assert (isinstance(strings, list) and isinstance(strings[0], str)), "input must be list of strings"
        strings = [simple_textcleaning(string,True) for string in strings]
        vectors = self.vectorize.transform(strings)
        results = self.xgb.predict(xgb.DMatrix(vectors),ntree_limit=self.xgb.best_ntree_limit)
        if get_proba:
            outputs = []
            for result in results:
                outputs.append({self.label[i]:result[i] for i in range(len(result))})
            return outputs
        else:
            return [self.label[i] for i in np.argmax(results,axis=1)]

class USER_BAYES:
    def __init__(self,multinomial, label, vectorize):
        self.multinomial = multinomial
        self.label = label
        self.vectorize = vectorize
    def predict(self, string, get_proba=False):
        assert (isinstance(string, str)), "input must be a string"
        vectors = self.vectorize.transform([simple_textcleaning(string,True)])
        if get_proba:
            result = self.multinomial.predict_proba(vectors)[0]
            return {self.label[i]:result[i] for i in range(len(result))}
        else:
            return self.label[self.multinomial.predict(vectors)[0]]
    def predict_batch(self, strings, get_proba=False):
        assert (isinstance(strings, list) and isinstance(strings[0], str)), "input must be list of strings"
        strings = [simple_textcleaning(string,True) for string in strings]
        vectors = self.vectorize.transform(strings)
        if get_proba:
            results = self.multinomial.predict_proba(vectors)
            outputs = []
            for result in results:
                outputs.append({self.label[i]:result[i] for i in range(len(result))})
            return outputs
        else:
            return [self.label[result] for result in self.multinomial.predict(vectors)]

def multinomial_detect_languages():
    if not os.path.isfile(bow_pkl):
        print('downloading LANGUAGE-DETECTION pickled bag-of-word')
        download_file("bow-language-detection.pkl", bow_pkl)
    if not os.path.isfile(multinomial_pkl):
        print('downloading LANGUAGE-DETECTION pickled multinomial model')
        download_file("multinomial-language-detection.pkl", multinomial_pkl)
    with open(bow_pkl,'rb') as fopen:
        BOW = pickle.load(fopen)
    with open(multinomial_pkl,'rb') as fopen:
        MULTINOMIAL = pickle.load(fopen)
    return USER_BAYES(MULTINOMIAL, lang_labels, BOW)

def xgb_detect_languages():
    if not os.path.isfile(xgb_bow_pkl):
        print('downloading LANGUAGE-DETECTION pickled bag-of-word XGB')
        download_file("bow-xgboost-language-detection.pkl", xgb_bow_pkl)
    if not os.path.isfile(xgb_pkl):
        print('downloading LANGUAGE-DETECTION pickled XGB model')
        download_file("xgboost-language-detection.pkl", xgb_pkl)
    with open(xgb_bow_pkl,'rb') as fopen:
        BOW = pickle.load(fopen)
    with open(xgb_pkl,'rb') as fopen:
        XGB = pickle.load(fopen)
    return USER_XGB(XGB, lang_labels, BOW)



from sklearn.utils import shuffle
from sklearn.manifold import TSNE
from scipy.spatial.distance import cdist
from sklearn.neighbors import NearestNeighbors
from fuzzywuzzy import fuzz
import pickle
import os
import sys
import collections
import re
import numpy as np
import tensorflow as tf
from . import home
from .utils import download_file

def malaya_word2vec(size = 256):
    if size not in [32,64,128,256,512]:
        raise Exception('size word2vec not supported')
    if not os.path.isfile('%s/word2vec-%d.p'%(home,size)):
        print('downloading word2vec-%d embedded'%(size))
        download_file('word2vec-%d.p'%(size),'%s/word2vec-%d.p'%(home,size))
    with open('%s/word2vec-%d.p'%(home,size), 'rb') as fopen:
        return pickle.load(fopen)

class Calculator():
    def __init__(self, tokens,):
        self._tokens = tokens
        self._current = tokens[0]

    def exp(self):
        result = self.term()
        while self._current in ('+', '-'):
            if self._current == '+':
                self.next()
                result += self.term()
            if self._current == '-':
                self.next()
                result -= self.term()
        return result

    def factor(self):
        result = None
        if self._current[0].isdigit() or self._current[-1].isdigit():
            result = np.array([float(i) for i in self._current.split(',')])
            self.next()
        elif self._current is '(':
            self.next()
            result = self.exp()
            self.next()
        return result

    def next(self):
        self._tokens = self._tokens[1:]
        self._current = self._tokens[0] if len(self._tokens) > 0 else None

    def term(self):
        result = self.factor()
        while self._current in ('*', '/'):
            if self._current == '*':
                self.next()
                result *= self.term()
            if self._current == '/':
                self.next()
                result /= self.term()
        return result

class Word2Vec:
    def __init__(self,embed_matrix, dictionary):
        self._embed_matrix = embed_matrix
        self._dictionary = dictionary
        self._reverse_dictionary = {v: k for k, v in dictionary.items()}
        self.words = list(dictionary.keys())

    def get_vector_by_name(self, word):
        return np.ravel(self._embed_matrix[self._dictionary[word], :])

    def calculator(self, equation, num_closest=5, metric='cosine', return_similarity=True):
        assert (isinstance(equation, str)), "input must be a string"
        tokens,temp = [], ''
        for char in equation:
            if char == ' ':
                continue
            if char not in '()*+-':
                temp += char
            else:
                if len(temp):
                    row = self._dictionary[self.words[np.argmax([fuzz.ratio(temp, k) for k in self.words])]]
                    tokens.append(','.join(self._embed_matrix[row,:].astype('str').tolist()))
                    temp = ''
                tokens.append(char)
        if len(temp):
            row = self._dictionary[self.words[np.argmax([fuzz.ratio(temp, k) for k in self.words])]]
            tokens.append(','.join(self._embed_matrix[row,:].astype('str').tolist()))
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1,metric=metric).fit(self._embed_matrix)
            distances, idx = nn.kneighbors(Calculator(tokens).exp().reshape((1,-1)))
            word_list = []
            for i in range(1,idx.shape[1]):
                word_list.append([self._reverse_dictionary[idx[0,i]],1-distances[0,i]])
            return word_list
        else:
            closest_indices = self.closest_row_indices(Calculator(tokens).exp(), num_closest + 1, metric)
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            return word_list

    def n_closest(self, word, num_closest=5, metric='cosine', return_similarity=True):
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1,metric=metric).fit(self._embed_matrix)
            distances, idx = nn.kneighbors(self._embed_matrix[self._dictionary[word], :].reshape((1,-1)))
            word_list = []
            for i in range(1,idx.shape[1]):
                word_list.append([self._reverse_dictionary[idx[0,i]],1-distances[0,i]])
            return word_list
        else:
            wv = self.get_vector_by_name(word)
            closest_indices = self.closest_row_indices(wv, num_closest + 1, metric)
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            if word in word_list:
                word_list.remove(word)
            return word_list

    def closest_row_indices(self, wv, num, metric):
        dist_array = np.ravel(cdist(self._embed_matrix, wv.reshape((1, -1)),metric=metric))
        sorted_indices = np.argsort(dist_array)
        return sorted_indices[:num]

    def analogy(self, a, b, c, num=1, metric='cosine'):
        va = self.get_vector_by_name(a)
        vb = self.get_vector_by_name(b)
        vc = self.get_vector_by_name(c)
        vd = vb - va + vc
        closest_indices = self.closest_row_indices(vd, num, metric)
        d_word_list = []
        for i in closest_indices:
            d_word_list.append(self._reverse_dictionary[i])
        return d_word_list

    def project_2d(self, start, end):
        tsne = TSNE(n_components=2)
        embed_2d = tsne.fit_transform(self._embed_matrix[start:end, :])
        word_list = []
        for i in range(start, end):
            word_list.append(self._reverse_dictionary[i])
        return embed_2d, word_list



from nltk.tokenize import word_tokenize
import os
import re
import pickle
import json
import xgboost as xgb
import numpy as np
from . import home
from .tatabahasa import tatabahasa_dict, hujung, permulaan
from .utils import download_file
from .text_functions import entities_textcleaning
from .keras_model import get_char_bidirectional, get_crf_lstm_bidirectional, get_crf_lstm_concat_bidirectional
from .keras_model import CONCAT_MODEL, WORD_MODEL, CHAR_MODEL

bow_pkl = home+'/bow-pos.pkl'
multinomial_pkl = home+'/multinomial-pos.pkl'
xgb_bow_pkl = home+'/xgb-bow-pos.pkl'
xgb_pkl = home+'/xgb-pos.pkl'

MULTINOMIAL, BOW = None, None
XGB, BOW_XGB = None, None

pos_labels = {0: 'ADJ', 1: 'ADP', 2: 'ADV', 3: 'AUX', 4: 'CCONJ', 5: 'DET',
6: 'NOUN', 7: 'NUM', 8: 'PART', 9: 'PRON', 10: 'PROPN',
11: 'SCONJ', 12: 'SYM', 13: 'VERB', 14: 'X'}

def naive_POS_word(word):
    for key, vals in tatabahasa_dict.items():
        if word in vals:
            return (key,word)
    try:
        if len(re.findall(r'^(.*?)(%s)$'%('|'.join(hujung[:1])), i)[0]) > 1:
            return ('KJ',word)
    except:
        pass
    try:
        if len(re.findall(r'^(.*?)(%s)'%('|'.join(permulaan[:-4])), word)[0]) > 1:
            return ('KJ',word)
    except Exception as e:
        pass
    if len(word) > 2:
        return ('KN',word)
    else:
        return ('',word)

def naive_pos(string):
    assert (isinstance(string, str)), "input must be a string"
    string = string.lower()
    results = []
    for i in word_tokenize(string):
        results.append(naive_POS_word(i))
    return results

def multinomial_pos(string):
    assert (isinstance(string, str)), "input must be a string"
    global MULTINOMIAL, BOW
    string = entities_textcleaning(string)
    if MULTINOMIAL is None and BOW is None:
        if not os.path.isfile(bow_pkl):
            print('downloading pickled bag-of-word multinomial POS')
            download_file("bow-entities.pkl", bow_pkl)
        if not os.path.isfile(multinomial_pkl):
            print('downloading pickled multinomial POS model')
            download_file("multinomial-entities.pkl", multinomial_pkl)
        with open(bow_pkl,'rb') as fopen:
            BOW = pickle.load(fopen)
        with open(multinomial_pkl,'rb') as fopen:
            MULTINOMIAL = pickle.load(fopen)
    return [(string[no],pos_labels[i]) for no, i in enumerate(MULTINOMIAL.predict(BOW.transform(string)))]

def xgb_pos(string):
    assert (isinstance(string, str)), "input must be a string"
    global XGB, BOW_XGB
    string = entities_textcleaning(string)
    if XGB is None and BOW_XGB is None:
        if not os.path.isfile(xgb_bow_pkl):
            print('downloading pickled bag-of-word XGB POS')
            download_file("xgb-bow-pos.pkl", xgb_bow_pkl)
        if not os.path.isfile(xgb_pkl):
            print('downloading pickled xgb POS model')
            download_file("xgb-pos.pkl", xgb_pkl)
        with open(xgb_bow_pkl,'rb') as fopen:
            BOW_XGB = pickle.load(fopen)
        with open(xgb_pkl,'rb') as fopen:
            XGB = pickle.load(fopen)
    dmatrix = xgb.DMatrix(BOW_XGB.transform(string))
    results = np.argmax(XGB.predict(dmatrix,ntree_limit=XGB.best_ntree_limit),axis=1)
    return [(string[no],pos_labels[i]) for no, i in enumerate(results)]

def get_available_pos_models():
    return ['char', 'word', 'concat']

char_frozen = home+'/pos-char-frozen.h5'
word_frozen = home+'/pos-word-frozen.h5'
concat_frozen = home+'/pos-concat-frozen.h5'

char_settings = home+'/pos-char-settings.json'
word_settings = home+'/pos-word-settings.json'
concat_settings = home+'/pos-concat-settings.json'

def deep_pos(model='concat'):
    if model == 'char':
        if not os.path.isfile(char_settings):
            print('downloading POS char settings')
            download_file("char-bidirectional-pos.json", char_settings)
        with open(char_settings,'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(char_frozen):
            print('downloading POS frozen char model')
            download_file("char-bidirectional-pos.h5", char_frozen)
        model = get_char_bidirectional(nodes['char2idx'],nodes['tag2idx'])
        model.load_weights(char_frozen)
        return CHAR_MODEL(model,nodes)
    elif model == 'word':
        if not os.path.isfile(word_settings):
            print('downloading POS word settings')
            download_file("crf-lstm-bidirectional-pos.json", word_settings)
        with open(word_settings,'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(word_frozen):
            print('downloading POS frozen word model')
            download_file("crf-lstm-bidirectional-pos.h5", word_frozen)
        model = get_crf_lstm_bidirectional(nodes['word2idx'],nodes['tag2idx'])
        model.load_weights(word_frozen)
        return WORD_MODEL(model,nodes)
    elif model == 'concat':
        if not os.path.isfile(concat_settings):
            print('downloading POS concat settings')
            download_file("crf-lstm-concat-bidirectional-pos.json", concat_settings)
        with open(concat_settings,'r') as fopen:
            nodes = json.loads(fopen.read())
            nodes['idx2word'] = {int(k):v for k,v in nodes['idx2word'].items()}
        if not os.path.isfile(concat_frozen):
            print('downloading POS frozen concat model')
            download_file("crf-lstm-concat-bidirectional-pos.h5", concat_frozen)
        model = get_crf_lstm_concat_bidirectional(nodes['char2idx'],nodes['word2idx'],nodes['tag2idx'])
        model.load_weights(concat_frozen)
        return CONCAT_MODEL(model,nodes)
    else:
        raise Exception('model not supported')



from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.decomposition import TruncatedSVD, NMF, LatentDirichletAllocation
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import Normalizer
from .text_functions import simple_textcleaning, STOPWORDS, classification_textcleaning, print_topics_modelling
from .stemmer import sastrawi_stemmer
from unidecode import unidecode
import itertools
import numpy as np
import re

class TOPIC:
    def __init__(self,features,comp,corpus,transformed):
        self.features = features
        self.comp = comp
        self.corpus = corpus
        self.transformed = transformed
    def print_topics(self, len_topic):
        print_topics_modelling(range(len_topic), feature_names = np.array(self.features),
        sorting = np.argsort(self.comp.components_)[:,::-1], topics_per_chunk=5, n_words=10)
    def get_topics(self, len_topic):
        results = []
        for no, topic in enumerate(self.comp.components_):
            results.append((no, " ".join([self.features[i] for i in topic.argsort()[:-len_topic -1:-1]])))
        return results
    def get_sentences(self, len_sentence, k=0):
        reverse_sorted = np.argsort(self.transformed[:,k])[::-1]
        return [self.corpus[i] for i in reverse_sorted[:len_sentence]]

def lda_topic_modelling(corpus,n_topics=10, max_df=0.95, min_df=2,
    stemming=True,cleaning=simple_textcleaning,stop_words=STOPWORDS):
    assert (isinstance(corpus, list) and isinstance(corpus[0], str)), "input must be list of strings"
    if cleaning is not None:
        for i in range(len(corpus)): corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)): corpus[i] = sastrawi_stemmer(corpus[i])
    tf_vectorizer = CountVectorizer(max_df=max_df, min_df=min_df, stop_words=stop_words)
    tf = tf_vectorizer.fit_transform(corpus)
    tf_features = tf_vectorizer.get_feature_names()
    lda = LatentDirichletAllocation(n_topics=n_topics, max_iter = 5, learning_method = 'online', learning_offset=50., random_state=0).fit(tf)
    return TOPIC(tf_features,lda,[classification_textcleaning(c) for c in corpus],lda.transform(tf))

def nmf_topic_modelling(corpus,n_topics=10, max_df=0.95, min_df=2,
    stemming=True,cleaning=simple_textcleaning,stop_words=STOPWORDS):
    assert (isinstance(corpus, list) and isinstance(corpus[0], str)), "input must be list of strings"
    if cleaning is not None:
        for i in range(len(corpus)): corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)): corpus[i] = sastrawi_stemmer(corpus[i])
    tfidf_vectorizer = TfidfVectorizer(max_df = max_df, min_df = min_df, stop_words = stop_words)
    tfidf = tfidf_vectorizer.fit_transform(corpus)
    tfidf_features = tfidf_vectorizer.get_feature_names()
    nmf = NMF(n_components=n_topics, random_state = 1, alpha =.1, l1_ratio=.5, init = 'nndsvd').fit(tfidf)
    return TOPIC(tfidf_features,nmf,[classification_textcleaning(c) for c in corpus],nmf.transform(tfidf))

def lsa_topic_modelling(corpus,n_topics, max_df=0.95, min_df=2,
    stemming=True,cleaning=simple_textcleaning,stop_words=STOPWORDS):
    assert (isinstance(corpus, list) and isinstance(corpus[0], str)), "input must be list of strings"
    if cleaning is not None:
        for i in range(len(corpus)): corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)): corpus[i] = sastrawi_stemmer(corpus[i])
    tfidf_vectorizer = TfidfVectorizer(max_df = max_df, min_df = min_df, stop_words = stop_words)
    tfidf = tfidf_vectorizer.fit_transform(corpus)
    tfidf_features = tfidf_vectorizer.get_feature_names()
    tfidf = Normalizer().fit_transform(tfidf)
    lsa = TruncatedSVD(n_topics).fit(tfidf)
    return TOPIC(tfidf_features,lsa,[classification_textcleaning(c) for c in corpus],lsa.transform(tfidf))



from sklearn.naive_bayes import MultinomialNB
from sklearn.preprocessing import LabelEncoder
from sklearn.feature_extraction.text import CountVectorizer, TfidfVectorizer
from sklearn.cross_validation import train_test_split
from sklearn import metrics, datasets
from keras.models import load_model
import tensorflow as tf
import numpy as np
import os
import json
from unidecode import unidecode
import pickle
import random
from .text_functions import separate_dataset, classification_textcleaning, str_idx, add_ngram, fasttext_str_idx
from .stemmer import naive_stemmer
from .language_detection import USER_XGB, USER_BAYES
from .utils import download_file, load_graph
from . import home
from .keras_model import CLASSIFIER

path_stack = home+'/stack-sentiment.hdf5'
path_stack_setting = home+'/stack-sentiment.json'
path_hierarchical = home+'/hierarchical-sentiment.pb'
path_hierarchical_setting = home+'/hierarchical-sentiment.json'
path_bahdanau = home+'/bahdanau-sentiment.pb'
path_bahdanau_setting = home+'/bahdanau-sentiment.json'
path_luong = home+'/luong-sentiment.pb'
path_luong_setting = home+'/luong-sentiment.json'
path_bidirectional = home+'/bidirectional-sentiment.pb'
path_bidirectional_setting = home+'/bidirectional-sentiment.json'
path_fasttext = home+'/fasttext-sentiment.pb'
path_fasttext_setting = home+'/fasttext-sentiment.json'
path_fasttext_pickle = home+'/fasttext-sentiment.pkl'
bayes_location = home + '/bayes-news.pkl'
tfidf_location = home + '/tfidf-news.pkl'
xgb_location = home + '/xgboost-sentiment.pkl'
xgb_tfidf_location = home + '/xgboost-tfidf.pkl'

class SENTIMENT:
    def __init__(self, X, logits, sess, mode, dictionary, ngram=None, alphas=None):
        self._X = X
        self._logits = logits
        self._sess = sess
        self._mode = mode
        self._dictionary = dictionary
        self._ngram = ngram
        self._alphas = alphas

    def predict(self,string):
        assert (isinstance(string, str)), 'input must be a string'
        string = classification_textcleaning(string,True)
        splitted = string.split()
        if self._mode == 'fast-text':
            batch_x = fasttext_str_idx([string], self._dictionary)
            batch_x = add_ngram(batch_x, self._ngram)
        else:
            batch_x = str_idx([string], self._dictionary, len(splitted), UNK=3)
        if self._mode in ['luong', 'bahdanau', 'hierarchical']:
            probs, alphas = self._sess.run([tf.nn.softmax(self._logits),self._alphas],feed_dict={self._X:batch_x})
            if self._mode == 'hierarchical':
                alphas = alphas[0]
            words = []
            for i in range(alphas.shape[0]):
                words.append([splitted[i],alphas[i]])
            return {'negative':probs[0,0],'positive':probs[0,1],'attention':words}
        if self._mode in ['bidirectional','fast-text']:
            probs = self._sess.run(tf.nn.softmax(self._logits),feed_dict={self._X:batch_x})
            return {'negative':probs[0,0],'positive':probs[0,1]}

    def predict_batch(self,strings):
        assert (isinstance(strings, list) and isinstance(strings[0], str)), 'input must be list of strings'
        strings = [classification_textcleaning(i,True) for i in strings]
        maxlen = max([len(i.split()) for i in strings])
        if self._mode == 'fast-text':
            batch_x = fasttext_str_idx(strings, self._dictionary)
            batch_x = add_ngram(batch_x, self._ngram)
            batch_x = tf.keras.preprocessing.sequence.pad_sequences(batch_x, maxlen)
        else:
            batch_x = str_idx(strings, self._dictionary, maxlen, UNK=3)
        probs = self._sess.run(tf.nn.softmax(self._logits),feed_dict={self._X:batch_x})
        dicts = []
        for i in range(probs.shape[0]):
            dicts.append({'negative':probs[i,0],'positive':probs[i,1]})
        return dicts

def get_available_sentiment_models():
    return ['bahdanau','hierarchical','luong','bidirectional','fast-text','stack']

def deep_sentiment(model='luong'):
    if model == 'stack':
        if not os.path.isfile(path_stack):
            print('downloading SENTIMENT frozen stack model')
            download_file('v6/stack-sentiment.hdf5',path_stack)
        if not os.path.isfile(path_stack_setting):
            print('downloading SENTIMENT stack dictionary')
            download_file('v6/stack-sentiment.json',path_stack_setting)
        with open(path_stack_setting,'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        return CLASSIFIER(load_model(path_stack),dictionary,['negative','positive'])
    elif model == 'fast-text':
        if not os.path.isfile(path_fasttext):
            print('downloading SENTIMENT frozen fast-text model')
            download_file('v5/fasttext-sentiment.pb',path_fasttext)
        if not os.path.isfile(path_fasttext_setting):
            print('downloading SENTIMENT fast-text dictionary')
            download_file('v5/fasttext-sentiment.json',path_fasttext_setting)
        if not os.path.isfile(path_fasttext_pickle):
            print('downloading SENTIMENT fast-text pickle')
            download_file('v5/fasttext-sentiment.pkl',path_fasttext_pickle)
        with open(path_fasttext_setting,'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        with open(path_fasttext_pickle, 'rb') as fopen:
            ngram = pickle.load(fopen)
        g=load_graph(path_fasttext)
        return SENTIMENT(g.get_tensor_by_name('import/Placeholder:0'),
                         g.get_tensor_by_name('import/logits:0'),
                         tf.InteractiveSession(graph=g),
                         model,
                         dictionary,
                         ngram=ngram)
    elif model == 'hierarchical':
        if not os.path.isfile(path_hierarchical):
            print('downloading SENTIMENT frozen hierarchical model')
            download_file('v5/hierarchical-sentiment.pb',path_hierarchical)
        if not os.path.isfile(path_hierarchical_setting):
            print('downloading SENTIMENT hierarchical dictionary')
            download_file('v5/hierarchical-sentiment.json',path_hierarchical_setting)
        with open(path_hierarchical_setting,'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        g=load_graph(path_hierarchical)
        return SENTIMENT(g.get_tensor_by_name('import/Placeholder:0'),
                         g.get_tensor_by_name('import/logits:0'),
                         tf.InteractiveSession(graph=g),
                         model,
                         dictionary,
                         alphas=g.get_tensor_by_name('import/alphas:0'))
    elif model == 'bahdanau':
        if not os.path.isfile(path_bahdanau):
            print('downloading SENTIMENT frozen bahdanau model')
            download_file('v5/bahdanau-sentiment.pb',path_bahdanau)
        if not os.path.isfile(path_bahdanau_setting):
            print('downloading SENTIMENT bahdanau dictionary')
            download_file('v5/bahdanau-sentiment.json',path_bahdanau_setting)
        with open(path_bahdanau_setting,'r') as fopen:
            dictionary = json.load(fopen)['dictionary']
        g=load_graph(path_bahdanau)
        return SENTIMENT(g.get_tensor_by_name('import/Placeholder:0'),
                         g.get_tensor_by_name('import/logits:0'),
                         tf.InteractiveSession(graph=g),
                         model,
                         dictionary,
                         alphas=g.get_tensor_by_name('import/alphas:0'))
    elif model == 'luong':
        if not os.path.isfile(path_luong):
            print('downloading SENTIMENT frozen luong model')
            download_file('v5/luong-sentiment.pb',path_luong)
        if not os.path.isfile(path_luong_setting):
            print('downloading SENTIMENT luong dictionary')
            download_file('v5/luong-sentiment.json',path_luong_setting)
        with open(path_luong_setting,'r') as fopen:
            dictionary = json.load(fopen)
        g=load_graph(path_luong)
        return SENTIMENT(g.get_tensor_by_name('import/Placeholder:0'),
                         g.get_tensor_by_name('import/logits:0'),
                         tf.InteractiveSession(graph=g),
                         model,
                         dictionary,
                         alphas=g.get_tensor_by_name('import/alphas:0'))
    elif model == 'bidirectional':
        if not os.path.isfile(path_bidirectional):
            print('downloading SENTIMENT frozen bidirectional model')
            download_file('v5/bidirectional-sentiment.pb',path_bidirectional)
        if not os.path.isfile(path_bidirectional_setting):
            print('downloading SENTIMENT bidirectional dictionary')
            download_file('v5/bidirectional-sentiment.json',path_bidirectional_setting)
        with open(path_bidirectional_setting,'r') as fopen:
            dictionary = json.load(fopen)
        g=load_graph(path_bidirectional)
        return SENTIMENT(g.get_tensor_by_name('import/Placeholder:0'),
                         g.get_tensor_by_name('import/logits:0'),
                         tf.InteractiveSession(graph=g),
                         model,
                         dictionary)
    else:
        raise Exception('model sentiment not supported')

def bayes_sentiment(
                corpus,
                cleaning = True,
                stemming = True,
                vector = 'tfidf',
                split_size = 0.2, **kwargs):
    multinomial,labels,vectorize = None, None, None
    if vector.lower().find('tfidf') < 0 and vector.lower().find('bow') < 0:
        raise Exception('Invalid vectorization technique')
    if isinstance(corpus, str):
        trainset = datasets.load_files(container_path = corpus, encoding = 'UTF-8')
        trainset.data, trainset.target = separate_dataset(trainset)
        data, target = trainset.data, trainset.target
        labels = trainset.target_names
    if isinstance(corpus, list) or isinstance(corpus, tuple):
        corpus = np.array(corpus)
        data, target = corpus[:,0].tolist(), corpus[:,1].tolist()
        labels = np.unique(target).tolist()
        target = LabelEncoder().fit_transform(target)
    c = list(zip(data, target))
    random.shuffle(c)
    data, target = zip(*c)
    data, target = list(data), list(target)
    if stemming:
        for i in range(len(data)): data[i] = ' '.join([naive_stemmer(k) for k in data[i].split()])
    if cleaning:
        for i in range(len(data)): data[i] = classification_textcleaning(data[i],True)
    if vector.lower().find('tfidf') >= 0:
        vectorize = TfidfVectorizer(**kwargs).fit(data)
        vectors = vectorize.transform(data)
    else:
        vectorize = CountVectorizer(**kwargs).fit(data)
        vectors = vectorize.transform(data)
    multinomial = MultinomialNB(**kwargs)
    if split_size:
        train_X, test_X, train_Y, test_Y = train_test_split(vectors, target, test_size = split_size)
        multinomial.partial_fit(train_X, train_Y,classes=np.unique(target))
        predicted = multinomial.predict(test_X)
        print(metrics.classification_report(test_Y, predicted, target_names = labels))
    else:
        multinomial.partial_fit(vectors,target,classes=np.unique(target))
        predicted = multinomial.predict(vectors)
        print(metrics.classification_report(target, predicted, target_names = labels))
    return USER_BAYES(multinomial, labels, vectorize)

def pretrained_bayes_sentiment():
    if not os.path.isfile(bayes_location):
        print('downloading SENTIMENT pickled multinomial model')
        download_file('v5/multinomial-sentiment.pkl', bayes_location)
    if not os.path.isfile(tfidf_location):
        print('downloading SENTIMENT pickled tfidf vectorizations')
        download_file('v5/tfidf-multinomial-sentiment.pkl', tfidf_location)
    with open(bayes_location,'rb') as fopen:
        multinomial = pickle.load(fopen)
    with open(tfidf_location,'rb') as fopen:
        vectorize = pickle.load(fopen)
    return USER_BAYES(multinomial, ['negative','positive'], vectorize)

def pretrained_xgb_sentiment():
    if not os.path.isfile(xgb_location):
        print('downloading SENTIMENT pickled XGB model')
        download_file('v5/xgboost-sentiment.pkl', xgb_location)
    if not os.path.isfile(xgb_tfidf_location):
        print('downloading SENTIMENT pickled tfidf vectorizations')
        download_file('v5/tfidf-xgboost-sentiment.pkl', xgb_tfidf_location)
    with open(xgb_location,'rb') as fopen:
        xgb = pickle.load(fopen)
    with open(xgb_tfidf_location,'rb') as fopen:
        vectorize = pickle.load(fopen)
    return USER_XGB(xgb, ['negative','positive'], vectorize)



from pathlib import Path
import shutil
import os
import warnings
import sys
warnings.simplefilter("ignore", DeprecationWarning)

home = str(Path.home())+'/Malaya'
version = '0.6'
version_path = home+'/version'

try:
    if not os.path.exists(home):
        os.makedirs(home)
except:
    print('cannot make directory for caching, exiting.')
    sys.exit(1)

if not os.path.isfile(version_path):
    print('not found any version, deleting previous version models..')
    shutil.rmtree(home)
    os.makedirs(home)
    with open(version_path,'w') as fopen:
        fopen.write(version)
else:
    with open(version_path,'r') as fopen:
        cached_version = fopen.read()
    if version not in cached_version:
        print('deleting previous version models..')
        shutil.rmtree(home)
        os.makedirs(home)
        with open(version_path,'w') as fopen:
            fopen.write(version)

from .entities import multinomial_entities, xgb_entities, deep_entities, get_available_entities_models
from .language_detection import multinomial_detect_languages, xgb_detect_languages, get_language_labels
from .normalizer import spell_normalizer, fuzzy_normalizer, load_malay_dictionary, basic_normalizer, deep_normalizer
from .num2word import to_cardinal,to_ordinal,to_ordinal_num,to_currency,to_year
from .pos_entities import deep_pos_entities, get_available_pos_entities_models
from .pos import naive_pos, multinomial_pos, xgb_pos, deep_pos, get_available_pos_models
from .sentiment import deep_sentiment, bayes_sentiment, pretrained_bayes_sentiment, get_available_sentiment_models, pretrained_xgb_sentiment
from .spell import naive_speller
from .stemmer import naive_stemmer, sastrawi_stemmer, deep_stemmer
from .summarization import summarize_lsa, summarize_nmf, summarize_lda
from .topic_modelling import lda_topic_modelling, nmf_topic_modelling, lsa_topic_modelling
from .topics_influencers import get_influencers, get_topics
from .toxic import multinomial_detect_toxic, logistics_detect_toxic, deep_toxic
from .word2vec import malaya_word2vec, Word2Vec

def describe_pos_malaya():
    print('KT - Kata Tanya')
    print('KJ - Kata Kerja')
    print('KP - Kata Perintah')
    print('KPA - Kata Pangkal')
    print('KB - Kata Bantu')
    print('KPENGUAT - Kata Penguat')
    print('KPENEGAS - Kata Penegas')
    print('NAFI - Kata Nafi')
    print('KPEMERI - Kata Pemeri')
    print('KS - Kata Sendi')
    print('KPEMBENAR - Kata Pembenar')
    print('NAFI - Kata Nafi')
    print('NO - Numbers')
    print('SUKU - Suku Bilangan')
    print('PISAHAN - Kata Pisahan')
    print('KETERANGAN - Kata Keterangan')
    print('ARAH - Kata Arah')
    print('KH - Kata Hubung')
    print('GN - Ganti Nama')
    print('KA - Kata Adjektif')
    print('O - not related, out scope')

def describe_pos():
    print('ADJ - Adjective, kata sifat')
    print('ADP - Adposition')
    print('ADV - Adverb, kata keterangan')
    print('ADX - Auxiliary verb, kata kerja tambahan')
    print('CCONJ - Coordinating conjuction, kata hubung')
    print('DET - Determiner, kata penentu')
    print('NOUN - Noun, kata nama')
    print('NUM - Number, nombor')
    print('PART - Particle')
    print('PRON - Pronoun, kata ganti')
    print('PROPN - Proper noun, kata ganti nama khas')
    print('SCONJ - Subordinating conjunction')
    print('SYM - Symbol')
    print('VERB - Verb, kata kerja')
    print('X - Other')

def describe_entities_malaya():
    print('PRN - person, group of people, believes, etc')
    print('LOC - location')
    print('NORP - Military, police, government, Parties, etc')
    print('ORG - Organization, company')
    print('LAW - related law document, etc')
    print('ART - art of work, special names, etc')
    print('EVENT - event happen, etc')
    print('FAC - facility, hospitals, clinics, etc')
    print('TIME - date, day, time, etc')
    print('O - not related, out scope')

def describe_entities():
    print('OTHER - Other')
    print('law - related law documents, documents, etc')
    print('location - location, place')
    print('organization - organization, company, government, facilities, etc')
    print('person - person, group of people, believes, etc')
    print('quantity - numbers, quantity')
    print('time - date, day, time, etc')



from collections import Counter, defaultdict
from fuzzywuzzy import fuzz
import numpy as np
from .text_functions import normalizer_textcleaning

alphabet = 'abcdefghijklmnopqrstuvwxyz'
consonants = 'bcdfghjklmnpqrstvwxyz'
vowels = 'aeiou'

def build_dicts(words):
    occurences = {}
    for l in alphabet:
        occurences[l] = defaultdict(lambda: 0)
    for word in words:
        occurences[word[0]][word] += 1
    return occurences

def edit_normalizer(word):
    splits = [(word[:i], word[i:]) for i in range(len(word) + 1)]
    deletes = [a + b[1:] for a, b in splits if b]
    transposes = [a + b[1] + b[0] + b[2:] for a, b in splits if len(b)>1]
    replaces = [a + c + b[1:] for a, b in splits for c in alphabet if b]
    inserts = [a + c + b for a, b in splits for c in alphabet]
    pseudo = []
    for c in vowels:
        pseudo.append(''.join([w+c for w in word]))
    fuzziness = []
    try:
        if word[-1] == 'e':
            fuzziness.append(word[:-1]+'a')
    except:
        pass
    try:
        if word[-2:] == 'ak':
            fuzziness.append(word[:-1])
    except:
        pass
    try:
        if word[0] in consonants and word[1] in consonants and word[2] in vowels:
            fuzziness.append(word[0]+word[2]+word[1:])
    except:
        pass
    try:
        if word[0] in vowels and word[-1] in consonants and word[2] in consonants:
            fuzziness.append(word[:-2]+word[0]+word[-1])
    except:
        pass
    try:
        if word[-1] == 'o' and word[-3] in vowels and word[-2] in consonants:
            fuzziness.append(word[:-1]+'ar')
    except:
        pass
    try:
        if word[0] == 'a' and word[1] in consonants:
            fuzziness.append('h'+word)
    except:
        pass
    return set(deletes + transposes + replaces + inserts + fuzziness + pseudo)

def return_possible(word, dicts, edits):
    return set(e2 for e1 in edits(word) for e2 in edits(e1) if e2 in dicts)

def return_known(word, dicts):
    return set(w for w in word if w in dicts)

class SPELL:
    def __init__(self,corpus):
        self.corpus = corpus
        self.occurences = build_dicts(self.corpus)
        self.corpus = Counter(corpus)

    def correct(self, string, first_char=True,debug=True):
        assert (isinstance(string, str)) and not string.count(' '), "input must be a single word"
        assert len(string) > 1, "input must be long than 1 characters"
        string = normalizer_textcleaning(string)
        if not string:
            return string
        if first_char:
            candidates = return_known([string], self.occurences[string[0]]) or return_known(edit_normalizer(string), self.occurences[string[0]]) or return_possible(string, self.occurences[string[0]], edit_normalizer) or [string]
        else:
            candidates = return_known([string], self.corpus) or return_known(edit_normalizer(string), self.corpus) or return_possible(string, self.corpus, edit_normalizer) or [string]
        candidates = list(candidates)
        if debug:
            print([(k,fuzz.ratio(string, k)) for k in candidates])
        return candidates[np.argmax([fuzz.ratio(string, k) for k in candidates])]

def naive_speller(corpus):
    assert (isinstance(corpus, list) and isinstance(corpus[0], str)), 'input must be list of strings'
    return SPELL(corpus)



import re
import tensorflow as tf
import numpy as np
import os
import json
from . import home
from .text_functions import process_word_pos_entities, char_str_idx, generate_char_seq
from .utils import load_graph, download_file

char_settings = home+'/char-settings.json'
char_frozen = home+'/char_frozen_model.pb'
concat_settings = home+'/concat-settings.json'
concat_frozen = home+'/concat_frozen_model.pb'
attention_settings = home+'/attention-settings.json'
attention_frozen = home+'/attention_pos_frozen_model.pb'

class DEEP_MODELS:
    def __init__(self,nodes,sess,predict):
        self.nodes = nodes
        self.sess = sess
        self.__predict = predict
    def predict(self,string):
        assert (isinstance(string, str)), "input must be a string"
        return self.__predict(string,self.sess,self.nodes)

def get_entity_char(string,sess,model):
    batch_x = char_str_idx([process_word_pos_entities(w) for w in string.split()],model['char2idx'],0)
    logits, logits_pos = sess.run([tf.argmax(model['logits'],1),tf.argmax(model['logits_pos'],1)],feed_dict={model['X']:batch_x})
    results = []
    for no, i in enumerate(string.split()):
        results.append((i,model['idx2tag'][str(logits[no])],model['idx2pos'][str(logits_pos[no])]))
    return results

def get_entity_concat(string,sess,model):
    array_X = char_str_idx([[process_word_pos_entities(w) for w in string.split()]],model['word2idx'],2)
    batch_x_char = generate_char_seq(array_X,model['idx2word'],model['char2idx'])
    Y_pred,Y_pos = sess.run([model['crf_decode'],model['crf_decode_pos']],feed_dict={model['word_ids']:array_X,
                                              model['char_ids']:batch_x_char})
    results = []
    for no, i in enumerate(string.split()):
        results.append((i,model['idx2tag'][str(Y_pred[0,no])],model['idx2pos'][str(Y_pos[0,no])]))
    return results

def get_available_pos_entities_models():
    return ['char', 'concat', 'attention']

def deep_pos_entities(model='attention'):
    if model == 'char':
        if not os.path.isfile(char_settings):
            print('downloading POS-ENTITIES char settings')
            download_file("char-settings-pos-entities.json", char_settings)
        with open(char_settings,'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(char_frozen):
            print('downloading POS-ENTITIES frozen char model')
            download_file("char-frozen-model-pos-entities.pb", char_frozen)
        g=load_graph(char_frozen)
        nodes['X'] = g.get_tensor_by_name('import/Placeholder:0')
        nodes['logits'] = g.get_tensor_by_name('import/logits:0')
        nodes['logits_pos'] = g.get_tensor_by_name('import/logits_pos:0')
        return DEEP_MODELS(nodes,tf.InteractiveSession(graph=g),get_entity_char)
    elif model == 'concat':
        if not os.path.isfile(concat_settings):
            print('downloading POS-ENTITIES concat settings')
            download_file("concat-settings-pos-entities.json", concat_settings)
        with open(concat_settings,'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(concat_frozen):
            print('downloading POS-ENTITIES frozen concat model')
            download_file("concat-frozen-model-pos-entities.pb", concat_frozen)
        g=load_graph(concat_frozen)
        nodes['word_ids'] = g.get_tensor_by_name('import/Placeholder:0')
        nodes['char_ids'] = g.get_tensor_by_name('import/Placeholder_1:0')
        nodes['crf_decode'] = g.get_tensor_by_name('import/entity-logits/cond/Merge:0')
        nodes['crf_decode_pos'] = g.get_tensor_by_name('import/pos-logits/cond/Merge:0')
        nodes['idx2word'] = {int(k):v for k,v in nodes['idx2word'].items()}
        return DEEP_MODELS(nodes,tf.InteractiveSession(graph=g),get_entity_concat)
    elif model == 'attention':
        if not os.path.isfile(attention_settings):
            print('downloading POS-ENTITIES attention settings')
            download_file("attention-settings-pos-entities.json", attention_settings)
        with open(attention_settings,'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(attention_frozen):
            print('downloading POS-ENTITIES frozen attention model')
            download_file("attention-frozen-model-pos-entities.pb", attention_frozen)
        g=load_graph(attention_frozen)
        nodes['word_ids'] = g.get_tensor_by_name('import/Placeholder:0')
        nodes['char_ids'] = g.get_tensor_by_name('import/Placeholder_1:0')
        nodes['crf_decode'] = g.get_tensor_by_name('import/entity-logits/cond/Merge:0')
        nodes['crf_decode_pos'] = g.get_tensor_by_name('import/pos-logits/cond/Merge:0')
        nodes['idx2word'] = {int(k):v for k,v in nodes['idx2word'].items()}
        return DEEP_MODELS(nodes,tf.InteractiveSession(graph=g),get_entity_concat)
    else:
        raise Exception('model not supported')



tanya_list = ['kenapa','bila','siapa','mengapa','apa','bagaimana','berapa','mana']
perintah_list = ['jangan','sila','tolong','harap','usah','jemput','minta']
pangkal_list = ['maka','alkisah','arakian','syahdah','adapun','bermula','kalakian']
bantu_list = ['akan','telah','boleh','mesti','belum','sudah','dapat','masih','harus','hendak']
penguat_list = ['paling','agak','sungguh','amat','terlalu','nian','benar','paling']
penegas_list = ['jua','juga','sahaja','hanya','memang','lagi','pun']
nafi_list = ['bukan','tidak','tak','tiada','tidaklah','tidakkah']
pemeri_list = ['ialah','adalah']
sendi_list = ['akan','kepada','terhadap','bagi','untuk','dari','daripada','di','dengan','hingga','sampai',
              'ke','kepada','oleh','pada','sejak','seperti','umpama','bak','tentang','laksanabagai',
             'semenjak','dalam','antara']
pembenar_list = ['ya','benar','betul']
nombor_list = ['satu','dua','tiga','empat','lima','enam','tujuh','lapan','sembilan','kosong']
suku_bilangan_list = ['per','suku','setengah','separuh','tiga suku']
pisahan_list = ['setiap','tiap']
keterangan_list = ['begitu','begini','demikian','perlahan','cepat','lena','akan','sedang','belum',
                   'telah','sekarang','sebentar','semalam','mungkin','agak','barangkali','pasti','tentu',
                   'sudah','selalu','kadang','acapkali','sesekali','yang']
arah_list = ['atas','bawah','tepi','antara','hadapan','utara','sisi','luar']
hubung_list = ['agar','apabila','atau','bahawa','dan','hingga','jika','jikalau','kecuali','kerana',
               'lalu','manakala','sambil','serta','semenjak','sementara','sungguhpun','supaya','walaupun','tetapi','berkenan','berkenaan']
gantinama_list = ['aku','saya','hamba','patik','beta','kami','kita','anda','awak','engkau','tuanku','kalian',
                  'kamu','baginda','beliau','mereka','ini','itu','sini','situ','sana','kini','dia','rm']

# pos permulaan[:-4]
permulaan = ['bel','be','se','ter','men','memper','di','pe','me','ke','ber','pen','per']
# pos hujung [:1]
hujung = ['kan', 'kah','lah','tah','nya','an','wan','wati','ita']
alphabet = 'qwertyuiopasdfghjklzxcvbnm'

tatabahasa_dict = {'KT':tanya_list,'KP':perintah_list,'KPA':pangkal_list,'KB':bantu_list,'KPENGUAT':penguat_list,
                   'KPENEGAS':penegas_list,'NAFI':nafi_list, 'KPEMERI':pemeri_list,'KS':sendi_list,'KPEMBENAR':pembenar_list,
                   'NO':nombor_list,'SUKU':suku_bilangan_list,'PISAHAN':pisahan_list,'KETERANGAN':keterangan_list,
                   'ARAH':arah_list,'KH':hubung_list,'GN':gantinama_list}

stopword_tatabahasa = list(set(tanya_list+perintah_list+pangkal_list+bantu_list+penguat_list+\
                penegas_list+nafi_list+pemeri_list+sendi_list+pembenar_list+nombor_list+\
                suku_bilangan_list+pisahan_list+keterangan_list+arah_list+hubung_list+gantinama_list))

rules_normalizer = {'experience': 'pengalaman',
 'bagasi': 'bagasi',
 'kg': 'kampung',
 'kilo': 'kilogram',
 'g': 'gram',
 'grm': 'gram',
 'k': 'okay',
 'abgkat': 'abang dekat',
 'abis': 'habis',
 'ade': 'ada',
 'adoi': 'aduh',
 'adoii': 'aduhh',
 'aerodarat': 'kapal darat',
 'agkt': 'angkat',
 'ahh': 'ah',
 'ailior': 'air liur',
 'airasia': 'air asia x',
 'airasiax': 'penerbangan',
 'airline': 'penerbangan',
 'airlines': 'penerbangan',
 'airport': 'lapangan terbang',
 'airpot': 'lapangan terbang',
 'aje': 'sahaja',
 'ajelah': 'sahajalah',
 'ajer': 'sahaja',
 'ak': 'aku',
 'aq': 'aku',
 'all': 'semua',
 'ambik': 'ambil',
 'amek': 'ambil',
 'amer': 'amir',
 'amik': 'ambil',
 'ana': 'saya',
 'angkt': 'angkat',
 'anual': 'tahunan',
 'apapun': 'apa pun',
 'ape': 'apa',
 'arab': 'arab',
 'area': 'kawasan',
 'aritu': 'hari itu',
 'ask': 'tanya',
 'astro': 'astro',
 'at': 'pada',
 'attitude': 'sikap',
 'babi': 'khinzir',
 'back': 'belakang',
 'bag': 'beg',
 'bang': 'abang',
 'bangla': 'bangladesh',
 'banyk': 'banyak',
 'bard': 'pujangga',
 'bargasi': 'bagasi',
 'bawak': 'bawa',
 'bawanges': 'bawang',
 'be': 'jadi',
 'behave': 'berkelakuan baik',
 'belagak': 'berlagak',
 'berdisiplin': 'berdisplin',
 'berenti': 'berhenti',
 'beskal': 'basikal',
 'bff': 'rakan karib',
 'bg': 'bagi',
 'bgi': 'bagi',
 'biase': 'biasa',
 'big': 'besar',
 'bike': 'basikal',
 'bile': 'bila',
 'binawe': 'binatang',
 'bini': 'isteri',
 'bkn': 'bukan',
 'bla': 'bila',
 'blom': 'belum',
 'bnyak': 'banyak',
 'body': 'tubuh',
 'bole': 'boleh',
 'boss': 'bos',
 'bowling': 'boling',
 'bpe': 'berapa',
 'brand': 'jenama',
 'brg': 'barang',
 'briefing': 'taklimat',
 'brng': 'barang',
 'bro': 'abang',
 'bru': 'baru',
 'bruntung': 'beruntung',
 'bsikal': 'basikal',
 'btnggjwb': 'bertanggungjawab',
 'btul': 'betul',
 'buatlh': 'buatlah',
 'buh': 'letak',
 'buka': 'buka',
 'but': 'tetapi',
 'bwk': 'bawa',
 'by': 'dengan',
 'byr': 'bayar',
 'bz': 'sibuk',
 'camera': 'kamera',
 'camni': 'macam ini',
 'cane': 'macam mana',
 'cant': 'tak boleh',
 'carakerja': 'cara kerja',
 'care': 'jaga',
 'cargo': 'kargo',
 'cctv': 'kamera litar tertutup',
 'celako': 'celaka',
 'cer': 'cerita',
 'cheap': 'murah',
 'check': 'semak',
 'ciput': 'sedikit',
 'cite': 'cerita',
 'citer': 'cerita',
 'ckit': 'sikit',
 'ckp': 'cakap',
 'class': 'kelas',
 'cm': 'macam',
 'cmni': 'macam ini',
 'cmpak': 'campak',
 'committed': 'komited',
 'company': 'syarikat',
 'complain': 'aduan',
 'corn': 'jagung',
 'couldnt': 'tak boleh',
 'cr': 'cari',
 'crew': 'krew',
 'cube': 'cuba',
 'cuma': 'cuma',
 'curinyaa': 'curinya',
 'cust': 'pelanggan',
 'customer': 'pelanggan',
 'd': 'di',
 'da': 'dah',
 'dn': 'dan',
 'dahh': 'dah',
 'damaged': 'rosak',
 'dapek': 'dapat',
 'day': 'hari',
 'dazrin': 'dazrin',
 'dbalingnya': 'dibalingnya',
 'de': 'ada',
 'deep': 'dalam',
 'deliberately': 'sengaja',
 'depa': 'mereka',
 'dessa': 'desa',
 'dgn': 'dengan',
 'dh': 'dah',
 'didunia': 'di dunia',
 'diorang': 'mereka',
 'diorng': 'mereka',
 'direct': 'secara terus',
 'diving': 'junam',
 'dkt': 'dekat',
 'dlempar': 'dilempar',
 'dlm': 'dalam',
 'dlt': 'padam',
 'dlu': 'dulu',
 'done': 'siap',
 'dont': 'jangan',
 'dorg': 'mereka',
 'dpermudhkn': 'dipermudahkan',
 'dpt': 'dapat',
 'dr': 'dari',
 'dri': 'dari',
 'dsb': 'dan sebagainya',
 'dy': 'dia',
 'educate': 'mendidik',
 'ensure': 'memastikan',
 'everything': 'semua',
 'ewahh': 'wah',
 'expect': 'sangka',
 'fb': 'facebook',
 'fired': 'pecat',
 'first': 'pertama',
 'fkr': 'fikir',
 'flight': 'kapal terbang',
 'for': 'untuk',
 'free': 'percuma',
 'friend': 'kawan',
 'fyi': 'untuk pengetahuan anda',
 'gantila': 'gantilah',
 'gantirugi': 'ganti rugi',
 'gentlemen': 'lelaki budiman',
 'gerenti': 'jaminan',
 'gile': 'gila',
 'gk': 'juga',
 'gnti': 'ganti',
 'go': 'pergi',
 'gomen': 'kerajaan',
 'goment': 'kerajaan',
 'good': 'baik',
 'ground': 'tanah',
 'guarno': 'macam mana',
 'hampa': 'mereka',
 'hampeh': 'teruk',
 'hanat': 'jahanam',
 'handle': 'kawal',
 'handling': 'kawalan',
 'hanta': 'hantar',
 'haritu': 'hari itu',
 'hate': 'benci',
 'have': 'ada',
 'hawau': 'celaka',
 'henpon': 'telefon',
 'heran': 'hairan',
 'him': 'dia',
 'his': 'dia',
 'hmpa': 'mereka',
 'hntr': 'hantar',
 'hotak': 'otak',
 'hr': 'hari',
 'i': 'saya',
 'hrga': 'harga',
 'hrp': 'harap',
 'hu': 'sedih',
 'humble': 'merendah diri',
 'ibon': 'ikon',
 'ichi': 'inci',
 'idung': 'hidung',
 'if': 'jika',
 'ig': 'instagram',
 'iklas': 'ikhlas',
 'improve': 'menambah baik',
 'in': 'masuk',
 'isn t': 'tidak',
 'isyaallah': 'insyallah',
 'ja': 'sahaja',
 'japan': 'jepun',
 'jd': 'jadi',
 'je': 'saja',
 'jee': 'saja',
 'jek': 'saja',
 'jepun': 'jepun',
 'jer': 'saja',
 'jerr': 'saja',
 'jez': 'saja',
 'jg': 'juga',
 'jgk': 'juga',
 'jgn': 'jangan',
 'jgnla': 'janganlah',
 'jibake': 'celaka',
 'jjur': 'jujur',
 'job': 'kerja',
 'jobscope': 'skop kerja',
 'jogja': 'jogjakarta',
 'jpam': 'jpam',
 'jth': 'jatuh',
 'jugak': 'juga',
 'ka': 'ke',
 'kalo': 'kalau',
 'kalu': 'kalau',
 'kang': 'nanti',
 'kantoi': 'temberang',
 'kasi': 'beri',
 'kat': 'dekat',
 'kbye': 'ok bye',
 'kearah': 'ke arah',
 'kecik': 'kecil',
 'keja': 'kerja',
 'keje': 'kerja',
 'kejo': 'kerja',
 'keksongan': 'kekosongan',
 'kemana': 'ke mana',
 'kene': 'kena',
 'kenekan': 'kenakan',
 'kesah': 'kisah',
 'ketempat': 'ke tempat',
 'kije': 'kerja',
 'kijo': 'kerja',
 'kiss': 'cium',
 'kite': 'kita',
 'kito': 'kita',
 'kje': 'kerja',
 'kjr': 'kerja',
 'kk': 'okay',
 'kmi': 'kami',
 'kt': 'kat',
 'tlg': 'tolong',
 'kl': 'kuala lumpur',
 'klai': 'kalau',
 'klau': 'kalau',
 'klia': 'klia',
 'klo': 'kalau',
 'klu': 'kalau',
 'kn': 'kan',
 'knapa': 'kenapa',
 'kne': 'kena',
 'ko': 'kau',
 'kompom': 'sah',
 'korang': 'kamu semua',
 'korea': 'korea',
 'korg': 'kamu semua',
 'kot': 'mungkin',
 'krja': 'kerja',
 'ksalahan': 'kesalahan',
 'kta': 'kita',
 'kuar': 'keluar',
 'kut': 'mungkin',
 'la': 'lah',
 'laa': 'lah',
 'lahabau': 'celaka',
 'lahanat': 'celaka',
 'lainda': 'lain dah',
 'lak': 'pula',
 'last': 'akhir',
 'le': 'lah',
 'leader': 'ketua',
 'leave': 'pergi',
 'ler': 'lah',
 'less': 'kurang',
 'letter': 'surat',
 'lg': 'lagi',
 'lgi': 'lagi',
 'lngsong': 'langsung',
 'lol': 'hehe',
 'lorr': 'lah',
 'low': 'rendah',
 'lps': 'lepas',
 'luggage': 'bagasi',
 'lumbe': 'lumba',
 'lyak': 'layak',
 'maap': 'maaf',
 'maapkan': 'maafkan',
 'mahai': 'mahal',
 'mampos': 'mampus',
 'mart': 'kedai',
 'mau': 'mahu',
 'mcm': 'macam',
 'mcmtu': 'macam itu',
 'memerlukn': 'memerlukan',
 'mengembirakan': 'menggembirakan',
 'mengmbilnyer': 'mengambilnya',
 'mengtasi': 'mengatasi',
 'mg': 'memang',
 'mihak': 'memihak',
 'min': 'admin',
 'mingu': 'minggu',
 'mintak': 'minta',
 'mjtuhkn': 'menjatuhkan',
 'mkyong': 'mak yong',
 'mlibatkn': 'melibatkan',
 'mmg': 'memang',
 'mmnjang': 'memanjang',
 'mmpos': 'mampus',
 'mn': 'mana',
 'mna': 'mana',
 'mntak': 'minta',
 'mntk': 'minta',
 'mnyusun': 'menyusun',
 'mood': 'suasana',
 'most': 'paling',
 'mr': 'tuan',
 'msa': 'masa',
 'msia': 'malaysia',
 'mst': 'mesti',
 'mu': 'awak',
 'much': 'banyak',
 'muko': 'muka',
 'mum': 'emak',
 'n': 'dan',
 'nah': 'nah',
 'nanny': 'nenek',
 'napo': 'kenapa',
 'nati': 'nanti',
 'ngan': 'dengan',
 'ngn': 'dengan',
 'ni': 'ini',
 'nie': 'ini',
 'nii': 'ini',
 'nk': 'nak',
 'nmpk': 'nampak',
 'nye': 'nya',
 'ofis': 'pejabat',
 'ohh': 'oh',
 'oii': 'hoi',
 'one': 'satu',
 'online': 'dalam talian',
 'or': 'atau',
 'org': 'orang',
 'orng': 'orang',
 'otek': 'otak',
 'p': 'pergi',
 'paid': 'dah bayar',
 'palabana': 'kepala otak',
 'pasni': 'lepas ini',
 'passengers': 'penumpang',
 'passengger': 'penumpang',
 'pastu': 'lepas itu',
 'pd': 'pada',
 'pegi': 'pergi',
 'pekerje': 'pekerja',
 'pekrja': 'pekerja',
 'perabih': 'perabis',
 'perkerja': 'pekerja',
 'pg': 'pergi',
 'phuii': 'puih',
 'pikir': 'fikir',
 'pilot': 'juruterbang',
 'pk': 'fikir',
 'pkerja': 'pekerja',
 'pkerjaan': 'pekerjaan',
 'pki': 'pakai',
 'please': 'tolong',
 'pls': 'tolong',
 'pn': 'pun',
 'pnh': 'pernah',
 'pnt': 'penat',
 'pnya': 'punya',
 'pon': 'pun',
 'priority': 'keutamaan',
 'properties': 'harta benda',
 'ptugas': 'petugas',
 'pub': 'kelab malam',
 'pulak': 'pula',
 'puye': 'punya',
 'pwrcuma': 'percuma',
 'pyahnya': 'payahnya',
 'quality': 'kualiti',
 'quit': 'keluar',
 'ramly': 'ramly',
 'rege': 'harga',
 'reger': 'harga',
 'report': 'laporan',
 'resigned': 'meletakkan jawatan',
 'respect': 'hormat',
 'rizal': 'rizal',
 'rosak': 'rosak',
 'rosok': 'rosak',
 'rse': 'rasa',
 'sacked': 'buang',
 'sado': 'tegap',
 'salute': 'sanjung',
 'sam': 'sama',
 'same': 'sama',
 'samp': 'sampah',
 'sbb': 'sebab',
 'sbgai': 'sebagai',
 'sblm': 'sebelum',
 'sblum': 'sebelum',
 'sbnarnya': 'sebenarnya',
 'sbum': 'sebelum',
 'sdg': 'sedang',
 'sebb': 'sebab',
 'sebijik': 'sebiji',
 'see': 'lihat',
 'seen': 'dilihat',
 'selangor': 'selangor',
 'selfie': 'swafoto',
 'sempoi': 'cantik',
 'senaraihitam': 'senarai hitam',
 'seorg': 'seorang',
 'service': 'perkhidmatan',
 'sgt': 'sangat',
 'shared': 'kongsi',
 'shirt': 'kemeja',
 'shut': 'tutup',
 'sib': 'nasib',
 'skali': 'sekali',
 'sket': 'sikit',
 'sma': 'sama',
 'smoga': 'semoga',
 'smpoi': 'cantik',
 'sndiri': 'sendiri',
 'sndr': 'sendiri',
 'sndri': 'sendiri',
 'sne': 'sana',
 'so': 'jadi',
 'sop': 'tatacara pengendalian piawai',
 'sorang': 'seorang',
 'spoting': 'pembintikan',
 'sronok': 'seronok',
 'ssh': 'susah',
 'staff': 'staf',
 'standing': 'berdiri',
 'start': 'mula',
 'steady': 'mantap',
 'stiap': 'setiap',
 'stress': 'stres',
 'student': 'pelajar',
 'study': 'belajar',
 'studycase': 'kajian kes',
 'sure': 'pasti',
 'sykt': 'syarikat',
 'tah': 'entah',
 'taik': 'tahi',
 'takan': 'tak akan',
 'takat': 'setakat',
 'takde': 'tak ada',
 'takkan': 'tak akan',
 'taknak': 'tak nak',
 'tang': 'tentang',
 'tanggungjawab': 'bertanggungjawab',
 'taraa': 'sementara',
 'tau': 'tahu',
 'tbabit': 'terbabit',
 'team': 'pasukan',
 'terbaekk': 'terbaik',
 'teruknye': 'teruknya',
 'tgk': 'tengok',
 'that': 'itu',
 'thinking': 'fikir',
 'those': 'itu',
 'time': 'masa',
 'tk': 'tak',
 'tnggongjwb': 'tanggungjawab',
 'tngok': 'tengok',
 'tngu': 'tunggu',
 'to': 'kepada',
 'tosak': 'rosak',
 'tp': 'tapi',
 'tpi': 'tapi',
 'tpon': 'telefon',
 'transfer': 'pindah',
 'trgelak': 'tergelak',
 'ts': 'tan sri',
 'tstony': 'tan sri tony',
 'tu': 'itu',
 'tuh': 'itu',
 'tula': 'itulah',
 'umeno': 'umno',
 'unfortunately': 'malangnya',
 'unhappy': 'tidak gembira',
 'up': 'naik',
 'upkan': 'naikkan',
 'ur': 'awak',
 'utk': 'untuk',
 'very': 'sangat',
 'viral': 'tular',
 'vote': 'undi',
 'warning': 'amaran',
 'warranty': 'waranti',
 'wassap': 'whatsapp',
 'wat': 'apa',
 'weii': 'wei',
 'well': 'maklumlah',
 'win': 'menang',
 'with': 'dengan',
 'wt': 'buat',
 'x': 'tak',
 'tw': 'tahu',
 'ye': 'ya',
 'yee': 'ya',
 'yg': 'yang',
 'yng': 'yang',
 'you': 'awak',
 'your': 'awak',
 'sakai':'selekeh',
 'rmb': 'billion ringgit',
 'rmj': 'juta ringgit',
 'rmk': 'ribu ringgit',
 'rm': 'ringgit'}



from tqdm import tqdm
import tensorflow as tf
import requests
import numpy as np

def download_file(url, filename):
    r = requests.get('http://s3-ap-southeast-1.amazonaws.com/huseinhouse-storage/'+url, stream = True)
    total_size = int(r.headers['content-length'])
    with open(filename, 'wb') as f:
        for data in tqdm(iterable = r.iter_content(chunk_size = 1048576), total = total_size/1048576, unit = 'MB',unit_scale=True):
            f.write(data)

def load_graph(frozen_graph_filename):
    with tf.gfile.GFile(frozen_graph_filename, "rb") as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())
    with tf.Graph().as_default() as graph:
        tf.import_graph_def(graph_def)
    return graph



import json
import pickle
import os
from .utils import download_file
from . import home
from keras.models import load_model
from scipy.sparse import hstack
from .keras_model import CLASSIFIER
from .text_functions import classification_textcleaning, str_idx

path_stack = home+'/stack-toxic.hdf5'
path_stack_setting = home+'/stack-toxic.json'
path_logistics = home+'/logistics-toxic.pkl'
path_multinomials = home+'/multinomials-toxic.pkl'
path_vectorizer_setting = home+'/vectorizer-toxic.pkl'
class_names = ['toxic', 'severe_toxic', 'obscene', 'threat', 'insult', 'identity_hate']

class TOXIC:
    def __init__(self,models,vectors):
        self._models = models
        self._vectors = vectors

    def _stack(self, strings):
        char_features = self._vectors['char'].transform(strings)
        word_features = self._vectors['word'].transform(strings)
        return hstack([char_features, word_features])

    def predict(self, string, get_proba=False):
        assert (isinstance(string, str)), "input must be a string"
        stacked = self._stack([classification_textcleaning(string,True)])
        result = {}
        for no, label in enumerate(class_names):
            if get_proba:
                result[label] = self._models[no].predict_proba(stacked)[0,1]
            else:
                result[label] = self._models[no].predict(stacked)[0]
        return result

    def predict_batch(self, strings, get_proba=False):
        assert (isinstance(strings, list) and isinstance(strings[0], str)), "input must be list of strings"
        stacked = self._stack([classification_textcleaning(i,True) for i in strings])
        result = {}
        for no, label in enumerate(class_names):
            if get_proba:
                result[label] = self._models[no].predict_proba(stacked)[:,1].tolist()
            else:
                result[label] = self._models[no].predict(stacked).tolist()
        return result

def multinomial_detect_toxic():
    if not os.path.isfile(path_multinomials):
        print('downloading TOXIC pickled multinomial model')
        download_file('v6/multinomials-toxic.pkl', path_multinomials)
    if not os.path.isfile(path_vectorizer_setting):
        print('downloading TOXIC pickled tfidf vectorizations')
        download_file('v6/vectorizer-toxic.pkl', path_vectorizer_setting)
    with open(path_multinomials,'rb') as fopen:
        multinomial = pickle.load(fopen)
    with open(path_vectorizer_setting,'rb') as fopen:
        vectorize = pickle.load(fopen)
    return TOXIC(multinomial, vectorize)

def logistics_detect_toxic():
    if not os.path.isfile(path_logistics):
        print('downloading TOXIC pickled logistics model')
        download_file('v6/logistics-toxic.pkl', path_logistics)
    if not os.path.isfile(path_vectorizer_setting):
        print('downloading TOXIC pickled tfidf vectorizations')
        download_file('v6/vectorizer-toxic.pkl', path_vectorizer_setting)
    with open(path_logistics,'rb') as fopen:
        logistic = pickle.load(fopen)
    with open(path_vectorizer_setting,'rb') as fopen:
        vectorize = pickle.load(fopen)
    return TOXIC(logistic, vectorize)

def deep_toxic():
    if not os.path.isfile(path_stack):
        print('downloading TOXIC frozen stack model')
        download_file('v6/stack-toxic.hdf5',path_stack)
    if not os.path.isfile(path_stack_setting):
        print('downloading TOXIC stack dictionary')
        download_file('v6/stack-toxic.json',path_stack_setting)
    with open(path_stack_setting,'r') as fopen:
        dictionary = json.load(fopen)['dictionary']
    return CLASSIFIER(load_model(path_stack),dictionary,class_names)



import os
from distutils.core import setup
from distutils import dir_util
from distutils import sysconfig
import setuptools


__packagename__ = 'malaya'

setuptools.setup(
    name=__packagename__,
    packages=setuptools.find_packages(),
    version='0.6.0.4',
    description='Natural-Language-Toolkit for bahasa Malaysia, powered by Deep Learning.',
    author='huseinzol05',
    author_email='husein.zol05@gmail.com',
    url='https://github.com/DevconX/Malaya',
    download_url='https://github.com/DevconX/Malaya/archive/master.zip',
    keywords=['nlp', 'bm'],
    install_requires=[
          'xgboost==0.80',
          'sklearn',
          'scikit-learn==0.19.1',
          'requests',
          'fuzzywuzzy',
          'tqdm',
          'nltk',
          'unidecode',
          'numpy',
          'scipy',
          'python-levenshtein',
          'pandas',
          'tensorflow==1.5'
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Text Processing',
    ],
    long_description=open('readme-pypi.srt').read())



import pickle
import os
import json
import xgboost as xgb
import numpy as np
from .utils import download_file
from . import home
from .text_functions import entities_textcleaning
from .keras_model import get_char_bidirectional, get_crf_lstm_bidirectional, get_crf_lstm_concat_bidirectional
from .keras_model import CONCAT_MODEL, WORD_MODEL, CHAR_MODEL

bow_pkl = home+'/bow-entities.pkl'
multinomial_pkl = home+'/multinomial-entities.pkl'
xgb_bow_pkl = home+'/xgb-bow-entities.pkl'
xgb_pkl = home+'/xgb-entities.pkl'

MULTINOMIAL, BOW = None, None
XGB, BOW_XGB = None, None

entities_labels = {0:'OTHER', 1:'law', 2:'location', 3:'organization',
4:'person', 5:'quantity', 6:'time'}

def multinomial_entities(string):
    assert (isinstance(string, str)), "input must be a string"
    global MULTINOMIAL, BOW
    string = entities_textcleaning(string)
    if MULTINOMIAL is None and BOW is None:
        if not os.path.isfile(bow_pkl):
            print('downloading ENTITIES pickled bag-of-word multinomial')
            download_file("bow-entities.pkl", bow_pkl)
        if not os.path.isfile(multinomial_pkl):
            print('downloading ENTITIES pickled multinomial model')
            download_file("multinomial-entities.pkl", multinomial_pkl)
        with open(bow_pkl,'rb') as fopen:
            BOW = pickle.load(fopen)
        with open(multinomial_pkl,'rb') as fopen:
            MULTINOMIAL = pickle.load(fopen)
    return [(string[no],entities_labels[i]) for no, i in enumerate(MULTINOMIAL.predict(BOW.transform(string)))]

def xgb_entities(string):
    assert (isinstance(string, str)), "input must be a string"
    global XGB, BOW_XGB
    string = entities_textcleaning(string)
    if XGB is None and BOW_XGB is None:
        if not os.path.isfile(xgb_bow_pkl):
            print('downloading ENTITIES pickled bag-of-word XGB')
            download_file("xgb-bow-entities.pkl", xgb_bow_pkl)
        if not os.path.isfile(xgb_pkl):
            print('downloading ENTITIES pickled XGB model')
            download_file("xgb-entities.pkl", xgb_pkl)
        with open(xgb_bow_pkl,'rb') as fopen:
            BOW_XGB = pickle.load(fopen)
        with open(xgb_pkl,'rb') as fopen:
            XGB = pickle.load(fopen)
    dmatrix = xgb.DMatrix(BOW_XGB.transform(string))
    results = np.argmax(XGB.predict(dmatrix,ntree_limit=XGB.best_ntree_limit),axis=1)
    return [(string[no],entities_labels[i]) for no, i in enumerate(results)]

def get_available_entities_models():
    return ['char', 'word', 'concat']

char_frozen = home+'/entities-char-frozen.h5'
word_frozen = home+'/entities-word-frozen.h5'
concat_frozen = home+'/entities-concat-frozen.h5'

char_settings = home+'/entities-char-settings.json'
word_settings = home+'/entities-word-settings.json'
concat_settings = home+'/entities-concat-settings.json'

def deep_entities(model='concat'):
    if model == 'char':
        if not os.path.isfile(char_settings):
            print('downloading ENTITIES char settings')
            download_file("char-bidirectional.json", char_settings)
        with open(char_settings,'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(char_frozen):
            print('downloading ENTITIES frozen char model')
            download_file("char-bidirectional.h5", char_frozen)
        model = get_char_bidirectional(nodes['char2idx'],nodes['tag2idx'])
        model.load_weights(char_frozen)
        return CHAR_MODEL(model,nodes)
    elif model == 'word':
        if not os.path.isfile(word_settings):
            print('downloading ENTITIES word settings')
            download_file("crf-lstm-bidirectional.json", word_settings)
        with open(word_settings,'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(word_frozen):
            print('downloading ENTITIES frozen word model')
            download_file("crf-lstm-bidirectional.h5", word_frozen)
        model = get_crf_lstm_bidirectional(nodes['word2idx'],nodes['tag2idx'])
        model.load_weights(word_frozen)
        return WORD_MODEL(model,nodes)
    elif model == 'concat':
        if not os.path.isfile(concat_settings):
            print('downloading ENTITIES concat settings')
            download_file("crf-lstm-concat-bidirectional.json", concat_settings)
        with open(concat_settings,'r') as fopen:
            nodes = json.loads(fopen.read())
            nodes['idx2word'] = {int(k):v for k,v in nodes['idx2word'].items()}
        if not os.path.isfile(concat_frozen):
            print('downloading ENTITIES frozen concat model')
            download_file("crf-lstm-concat-bidirectional.h5", concat_frozen)
        model = get_crf_lstm_concat_bidirectional(nodes['char2idx'],nodes['word2idx'],nodes['tag2idx'])
        model.load_weights(concat_frozen)
        return CONCAT_MODEL(model,nodes)
    else:
        raise Exception('model not supported')



import keras
import numpy as np
import tensorflow as tf
from keras.models import Model, Input
from keras.layers import LSTM, Embedding, Dense, TimeDistributed, Dropout, Bidirectional, Reshape, Concatenate, Lambda
from keras_contrib.layers import CRF
from keras import backend as K
from keras.backend.tensorflow_backend import set_session
from .text_functions import entities_textcleaning, classification_textcleaning, char_str_idx, generate_char_seq, str_idx
set_session(tf.InteractiveSession())

class CLASSIFIER:
    def __init__(self, model, dictionary, labels):
        self._model = model
        self._dictionary = dictionary
        self._labels = labels

    def predict(self, string):
        assert (isinstance(string, str)), "input must be a string"
        string = classification_textcleaning(string,True)
        splitted = string.split()
        batch_x = str_idx([string], self._dictionary, len(splitted), UNK=3)
        probs = self._model.predict(batch_x)
        return {label:probs[0,no] for no, label in enumerate(self._labels)}

    def predict_batch(self, strings):
        assert (isinstance(strings, list) and isinstance(strings[0], str)), 'input must be list of strings'
        strings = [classification_textcleaning(i,True) for i in strings]
        maxlen = max([len(i.split()) for i in strings])
        batch_x = str_idx(strings, self._dictionary, maxlen, UNK=3)
        probs = self._model.predict(batch_x)
        dicts = []
        for i in range(probs.shape[0]):
            dicts.append({label:probs[i,no] for no, label in enumerate(self._labels)})
        return dicts

class CONCAT_MODEL:
    def __init__(self, model, settings):
        self._model = model
        self._settings = settings
    def predict(self, string):
        assert (isinstance(string, str)), "input must be a string"
        string = entities_textcleaning(string.lower())
        batch_x = char_str_idx([string],self._settings['word2idx'],2)
        batch_x_char = generate_char_seq(batch_x,self._settings['idx2word'],self._settings['char2idx'])
        results = np.argmax(self._model.predict([batch_x, batch_x_char])[0],1)
        return [(string[no],self._settings['idx2tag'][str(i)]) for no, i in enumerate(results)]

class WORD_MODEL:
    def __init__(self, model, settings):
        self._model = model
        self._settings = settings
    def predict(self, string):
        assert (isinstance(string, str)), "input must be a string"
        string = entities_textcleaning(string.lower())
        batch_x = char_str_idx([string],self._settings['word2idx'],2)
        results = np.argmax(self._model.predict(batch_x)[0],1)
        return [(string[no],self._settings['idx2tag'][str(i)]) for no, i in enumerate(results)]

class CHAR_MODEL:
    def __init__(self, model, settings):
        self._model = model
        self._settings = settings
    def predict(self, string):
        assert (isinstance(string, str)), "input must be a string"
        string = entities_textcleaning(string.lower())
        batch_x = char_str_idx(string,self._settings['char2idx'],0)
        results = np.argmax(self._model.predict(batch_x),1)
        return [(string[no],self._settings['idx2tag'][str(i)]) for no, i in enumerate(results)]

def get_char_bidirectional(char2idx,tag2idx):
    input_word = Input(shape=(None,))
    model = Embedding(input_dim=len(char2idx) + 1, output_dim=128, mask_zero=True)(input_word)
    model = Bidirectional(LSTM(units=128, return_sequences=False, recurrent_dropout=0.1))(model)
    out = Dense(len(tag2idx),activation='softmax')(model)
    model = Model(input_word, out)
    return model

def get_crf_lstm_bidirectional(word2idx,tag2idx):
    input_word = Input(shape=(None,))
    model = Embedding(input_dim=len(word2idx) + 1, output_dim=128, mask_zero=True)(input_word)
    model = Bidirectional(LSTM(units=64, return_sequences=True, recurrent_dropout=0.1))(model)
    model = TimeDistributed(Dense(50, activation="relu"))(model)
    crf = CRF(len(tag2idx))
    out = crf(model)
    model = Model(input_word, out)
    return model

def get_crf_lstm_concat_bidirectional(char2idx,word2idx,tag2idx):
    input_word = Input(shape=(None,))
    input_char = Input(shape=(None,None,))
    model_char = Embedding(input_dim=len(char2idx) + 1, output_dim=128)(input_char)
    s = K.shape(model_char)
    def backend_reshape(x):
        return K.reshape(x, (s[0]*s[1],s[2],128))
    model_char = Lambda(backend_reshape)(model_char)
    model_char = Bidirectional(LSTM(units=50, return_sequences=True, recurrent_dropout=0.1))(model_char)
    def sliced(x):
        return x[:,-1]
    model_char = Lambda(sliced)(model_char)
    def backend_reshape(x):
        return K.reshape(x, (s[0],s[1],100))
    model_char = Lambda(backend_reshape)(model_char)
    model_word = Embedding(input_dim=len(word2idx) + 1, output_dim=64, mask_zero=True)(input_word)
    concated_word_char = Concatenate(-1)([model_char,model_word])
    model = Bidirectional(LSTM(units=50, return_sequences=True, recurrent_dropout=0.1))(concated_word_char)
    model = TimeDistributed(Dense(50, activation="relu"))(model)
    crf = CRF(len(tag2idx))
    output = crf(model)
    model = Model(inputs=[input_word, input_char], outputs=output)
    return model



import numpy as np
from fuzzywuzzy import fuzz
import pickle
import os
import json
import tensorflow as tf
from collections import Counter
from . import home
from .utils import load_graph, download_file
from .num2word import to_cardinal
from .text_functions import normalizer_textcleaning, stemmer_str_idx, pad_sentence_batch
from .tatabahasa import rules_normalizer
from .spell import return_possible, edit_normalizer, build_dicts, return_known

consonants = 'bcdfghjklmnpqrstvwxyz'
vowels = 'aeiou'
sounds = {'x':'tidak','y':'kenapa','n':'dan','g':'pergi','s':'seperti','d':'di','k':'ok','u':'awak',
        't':'nanti','p':'pergi','wai':'kenapa'}
GO = 0
PAD = 1
EOS = 2
UNK = 3
malay_text = home+'/malay-text.txt'
normalizer_json = home+'/normalizer-deep.json'
normalizer_graph = home+'/normalizer-deep.pb'

class DEEP_NORMALIZER:
    def __init__(self, x, logits, sess, dicts):
        self._sess = sess
        self._x = x
        self._logits = logits
        self._dicts = dicts
        self._dicts['rev_dictionary_to'] = {int(k):v for k,v in self._dicts['rev_dictionary_to'].items()}

    def normalize(self,string):
        assert (isinstance(string, str)), "input must be a string"
        token_strings = normalizer_textcleaning(string).split()
        idx = stemmer_str_idx(token_strings,self._dicts['dictionary_from'])
        predicted = self._sess.run(self._logits, feed_dict={self._x:pad_sentence_batch(idx, PAD)[0]})
        results = []
        for word in predicted:
            results.append(''.join([self._dicts['rev_dictionary_to'][c] for c in word if c not in[GO,PAD,EOS,UNK]]))
        return ' '.join(results)

class SPELL_NORMALIZE:
    def __init__(self,corpus):
        self.corpus = Counter(corpus)

    def normalize(self, string, debug=True):
        assert (isinstance(string, str)), "input must be a string"
        result = []
        for word in normalizer_textcleaning(string).split():
            if word[0] == 'x' and len(word) > 1:
                result_string = 'tak '
                word = word[1:]
            else:
                result_string = ''
            if word[-2:] == 'la':
                end_result_string = ' lah'
                word = word[:-2]
            elif word[-3:] == 'lah':
                end_result_string = ' lah'
                word = word[:-3]
            else:
                end_result_string = ''
            if word in sounds:
                result.append(result_string + sounds[word] + end_result_string)
                continue
            if word in rules_normalizer:
                result.append(result_string + rules_normalizer[word] + end_result_string)
                continue
            if word in self.corpus:
                result.append(result_string + word + end_result_string)
                continue
            candidates = return_known([word], self.corpus) or return_known(edit_normalizer(word), self.corpus) or return_possible(word, self.corpus, edit_normalizer) or [word]
            candidates = list(candidates)
            if debug:
                print([(k,fuzz.ratio(word, k)) for k in candidates])
            result.append(result_string + candidates[np.argmax([fuzz.ratio(word, k) for k in candidates])] + end_result_string)
        return ' '.join(result)

class FUZZY_NORMALIZE:
    def __init__(self,normalized,corpus):
        self.normalized = normalized
        self.corpus = corpus

    def normalize(self,string):
        assert (isinstance(string, str)), "input must be a string"
        result = []
        for word in normalizer_textcleaning(string).split():
            if word[0] == 'x' and len(word) > 1:
                result_string = 'tak '
                word = word[1:]
            else:
                result_string = ''
            if word[-2:] == 'la':
                end_result_string = ' lah'
                word = word[:-2]
            elif word[-3:] == 'lah':
                end_result_string = ' lah'
                word = word[:-3]
            else:
                end_result_string = ''
            if word in sounds:
                result.append(result_string + sounds[word] + end_result_string)
                continue
            if word in rules_normalizer:
                result.append(result_string + rules_normalizer[word] + end_result_string)
                continue
            if word in self.corpus:
                result.append(result_string + word + end_result_string)
                continue
            results = []
            for i in range(len(self.normalized)):
                results.append(np.mean([fuzz.ratio(word, k) for k in self.normalized[i]]))
            if len(np.where(np.array(results) > 70)[0]) < 1:
                result.append(result_string + word + end_result_string)
                continue
            result.append(result_string + self.corpus[np.argmax(results)] + end_result_string)
        return ' '.join(result)

def fuzzy_normalizer(corpus):
    assert (isinstance(corpus, list) and isinstance(corpus[0], str)), 'input must be list of strings'
    transform = []
    for i in corpus:
        i = i.lower()
        result = []
        result.append(i)
        result.append(''.join(char for char in i if char not in vowels))
        if i[0] in consonants and i[-1] in consonants:
            result.append(i[0]+i[-1])
        if i[-1] == 'a':
            result.append(i[:-1]+'e')
            result.append(i+'k')
        if i[1] in vowels and i[0] in consonants:
            result.append(i[0]+i[2:])
        if i[-2] in vowels and i[-1] in consonants:
            result.append(i[:-2]+i[-1])
        result.append(i[0]+i[-1])
        if i[-2:] == 'ar':
            result.append(i[:-2]+'o')
        if i[:2] == 'ha':
            result.append(i[1:])
        transform.append(list(set(result)))
    return FUZZY_NORMALIZE(transform,corpus)

def spell_normalizer(corpus):
    assert (isinstance(corpus, list) and isinstance(corpus[0], str)), 'input must be list of strings'
    return SPELL_NORMALIZE(corpus)

def load_malay_dictionary():
    if not os.path.isfile(malay_text):
        print('downloading Malay texts')
        download_file('v6/malay-text.txt', malay_text)
    with open(malay_text,'r') as fopen:
        return [text.lower() for text in (list(filter(None, fopen.read().split('\n'))))]

def basic_normalizer(string):
    assert (isinstance(string, str)), "input must be a string"
    result = []
    for i in normalizer_textcleaning(string).split():
        if i in sounds:
            result.append(sounds[i])
        elif i[-1] == '2':
            result.append(i[:-1])
        else:
            result.append(i)
    return ' '.join(result)

def deep_normalizer():
    if not os.path.isfile(normalizer_json):
        print('downloading JSON normalizer')
        download_file("v6/normalizer-deep.json", normalizer_json)
    with open(normalizer_json,'r') as fopen:
        dic_normalizer = json.load(fopen)
    if not os.path.isfile(normalizer_graph):
        print('downloading normalizer graph')
        download_file("v6/normalizer-deep.pb", normalizer_graph)
    g=load_graph(normalizer_graph)
    return DEEP_NORMALIZER(g.get_tensor_by_name('import/Placeholder:0'),
                        g.get_tensor_by_name('import/logits:0'),
                        tf.InteractiveSession(graph=g),
                        dic_normalizer)



import numpy as np
from scipy.linalg import svd, inv
import re, random
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.decomposition import NMF, LatentDirichletAllocation
from operator import itemgetter
from .text_functions import cluster_words, summary_textcleaning, classification_textcleaning, STOPWORDS
from .stemmer import sastrawi_stemmer

def summarize_lsa(corpus, maintain_original = False, top_k = 3, important_words = 3,
    return_cluster = True, **kwargs):
    assert (isinstance(corpus, list) and isinstance(corpus[0], str)), "input must be list of strings"
    corpus = [summary_textcleaning(i) for i in corpus]
    corpus = ' '.join(corpus)
    splitted_fullstop = re.findall('(?=\S)[^.\n]+(?<=\S)', corpus)
    splitted_fullstop = [classification_textcleaning(i) if not maintain_original else i for i in splitted_fullstop if len(i)]
    stemmed = [sastrawi_stemmer(i) for i in splitted_fullstop]
    tfidf = TfidfVectorizer(ngram_range=(1, 3),min_df=2,stop_words=STOPWORDS,**kwargs).fit(stemmed)
    U, S, Vt = svd(tfidf.transform(stemmed).todense().T, full_matrices =False)
    summary =[(splitted_fullstop[i],
    np.linalg.norm(np.dot(np.diag(S),Vt[:,b]),2)) for i in range(len(splitted_fullstop)) for b in range(len(Vt))]
    summary = sorted(summary, key=itemgetter(1))
    summary = dict((v[0],v) for v in sorted(summary, key=lambda summary: summary[1])).values()
    summarized = '. '.join([a for a, b in summary][len(summary)-(top_k):])
    indices = np.argsort(tfidf.idf_)[::-1]
    features = tfidf.get_feature_names()
    top_words = [features[i] for i in indices[:important_words]]
    if return_cluster:
        return {'summary':summarized, 'top-words': top_words, 'cluster-top-words':cluster_words(top_words)}
    return {'summary':summarized, 'top-words': top_words}

def summarize_nmf(corpus, maintain_original = False, top_k = 3, important_words = 3,
    return_cluster = True, **kwargs):
    assert (isinstance(corpus, list) and isinstance(corpus[0], str)), "input must be list of strings"
    corpus = [summary_textcleaning(i) for i in corpus]
    corpus = ' '.join(corpus)
    splitted_fullstop = re.findall('(?=\S)[^.\n]+(?<=\S)', corpus)
    splitted_fullstop = [classification_textcleaning(i) if not maintain_original else i for i in splitted_fullstop if len(i)]
    stemmed = [sastrawi_stemmer(i) for i in splitted_fullstop]
    tfidf = TfidfVectorizer(ngram_range=(1, 3),min_df=2,stop_words=STOPWORDS,**kwargs).fit(stemmed)
    densed_tfidf = tfidf.transform(stemmed).todense()
    nmf = NMF(len(splitted_fullstop)).fit(densed_tfidf)
    vectors = nmf.transform(densed_tfidf)
    components = nmf.components_.mean(axis=1)
    summary =[(splitted_fullstop[i],
    np.linalg.norm(np.dot(np.diag(components),vectors[:,b]),2)) for i in range(len(splitted_fullstop)) for b in range(len(vectors))]
    summary = sorted(summary, key=itemgetter(1))
    summary = dict((v[0],v) for v in sorted(summary, key=lambda summary: summary[1])).values()
    summarized = '. '.join([a for a, b in summary][len(summary)-(top_k):])
    indices = np.argsort(tfidf.idf_)[::-1]
    features = tfidf.get_feature_names()
    top_words = [features[i] for i in indices[:important_words]]
    if return_cluster:
        return {'summary':summarized, 'top-words': top_words, 'cluster-top-words':cluster_words(top_words)}
    return {'summary':summarized, 'top-words': top_words}

def summarize_lda(corpus, maintain_original = False, top_k = 3, important_words = 3,
    return_cluster = True, **kwargs):
    assert (isinstance(corpus, list) and isinstance(corpus[0], str)), "input must be list of strings"
    corpus = [summary_textcleaning(i) for i in corpus]
    corpus = ' '.join(corpus)
    splitted_fullstop = re.findall('(?=\S)[^.\n]+(?<=\S)', corpus)
    splitted_fullstop = [classification_textcleaning(i) if not maintain_original else i for i in splitted_fullstop if len(i)]
    stemmed = [sastrawi_stemmer(i) for i in splitted_fullstop]
    tfidf = TfidfVectorizer(ngram_range=(1, 3),min_df=2,stop_words=STOPWORDS,**kwargs).fit(stemmed)
    densed_tfidf = tfidf.transform(stemmed).todense()
    lda = LatentDirichletAllocation(len(splitted_fullstop)).fit(densed_tfidf)
    vectors = lda.transform(densed_tfidf)
    components = lda.components_.mean(axis=1)
    summary =[(splitted_fullstop[i],
    np.linalg.norm(np.dot(np.diag(components),vectors[:,b]),2)) for i in range(len(splitted_fullstop)) for b in range(len(vectors))]
    summary = sorted(summary, key=itemgetter(1))
    summary = dict((v[0],v) for v in sorted(summary, key=lambda summary: summary[1])).values()
    summarized = '. '.join([a for a, b in summary][len(summary)-(top_k):])
    indices = np.argsort(tfidf.idf_)[::-1]
    features = tfidf.get_feature_names()
    top_words = [features[i] for i in indices[:important_words]]
    if return_cluster:
        return {'summary':summarized, 'top-words': top_words, 'cluster-top-words':cluster_words(top_words)}
    return {'summary':summarized, 'top-words': top_words}
